"""Module for experimental projection features. This API may change in the future."""

import _zivid
from zivid.frame_2d import Frame2D
from zivid.settings2d import Settings2D, _to_internal_settings2d


class ProjectedImage:
    """A handle to a 2D image being displayed on a Zivid camera's projector.

    The image projection will stop either when the instance is destroyed, when the stop() method is called,
    or when exiting the "with" block if used as a context manager.
    """

    def __init__(self, impl):
        """Initialize ProjectedImage wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl: Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if isinstance(impl, _zivid.ProjectedImage):
            self.__impl = impl
        else:
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}.".format(
                    type(impl), _zivid.ProjectedImage
                )
            )

    def __str__(self):
        return str(self.__impl)

    def capture(self, settings2d):
        """Capture a single 2D frame without stopping the ongoing image projection.

        This method returns right after the acquisition of the image is complete. This function can only be used
        with a zero-brightness 2D capture, otherwise it will interfere with the projected image. An exception
        will be thrown if settings contains brightness > 0.

        Args:
            settings2d: A Settings2D instance to be used for 2D capture.

        Returns:
            A Frame2D containing a 2D image plus metadata.

        Raises:
            TypeError: If argument is not a Settings2D.
        """

        if isinstance(settings2d, Settings2D):
            return Frame2D(self.__impl.capture(_to_internal_settings2d(settings2d)))
        raise TypeError("Unsupported settings type: {}".format(type(settings2d)))

    def stop(self):
        """Stop the ongoing image projection."""
        self.__impl.stop()

    def active(self):
        """Check if a handle is associated with an ongoing image projection.

        Returns:
            A boolean indicating projection state.
        """
        return self.__impl.active()

    def release(self):
        """Release the underlying resources and stop projection."""
        try:
            impl = self.__impl
        except AttributeError:
            pass
        else:
            impl.release()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.release()

    def __del__(self):
        self.release()


def projector_resolution(camera):
    """Get the resolution of the internal projector in the Zivid camera.

    Args:
        camera: The Camera instance to get the projector resolution of.

    Returns:
        The resolution as a tuple (height,width).
    """
    return _zivid.projection.projector_resolution(
        camera._Camera__impl  # pylint: disable=protected-access
    )


def show_image_bgra(camera, image_bgra):
    """Display a 2D color image using the projector.

    The image resolution needs to be the same as the resolution obtained from the projector_resolution
    function for the camera. This function returns a ProjectedImage instance. Projection will continue
    until this object is destroyed, if its stop() method is called, or if another capture is initialized
    on the same camera. The ProjectedImage object can be used as a context manager, in which case the
    projection will stop when exiting the "with" block.

    Args:
        camera: The Camera instance to project with.
        image_bgra: The image to project in the form of a HxWx4 numpy array with BGRA colors.

    Returns:
        A handle in the form of a ProjectedImage instance.
    """

    return ProjectedImage(
        _zivid.projection.show_image_bgra(
            camera._Camera__impl,  # pylint: disable=protected-access
            image_bgra,
        )
    )


def pixels_from_3d_points(camera, points):
    """Get 2D projector pixel coordinates corresponding to 3D points relative to the camera.

    This function takes 3D points in the camera's reference frame and converts them to the projector frame
    using the internal calibration of a Zivid camera. In a Zivid point cloud, each point corresponds to a
    pixel coordinate in the camera, but the projector has a slight offset. The translation of each point
    depends on the distance between the camera and the point, as well as the distance and angle between the
    camera and the projector.

    Args:
        camera: The Camera instance that the 3D points are in the frame of.
        points: A list of 3D (XYZ) points as List[List[float[3]]] or Nx3 Numpy array.

    Returns:
        The corresponding 2D (XY) points in the projector (List[List[float[2]]])
    """

    return _zivid.projection.pixels_from_3d_points(
        camera._Camera__impl,  # pylint: disable=protected-access
        points,
    )



import pytest


def test_init_default_settings(application):
    import zivid

    settings_2d = zivid.Settings2D()

    assert isinstance(settings_2d.acquisitions, list)
    assert len(settings_2d.acquisitions) == 0
    assert isinstance(settings_2d.processing, zivid.Settings2D.Processing)

    assert isinstance(settings_2d.processing.color, zivid.Settings2D.Processing.Color)
    assert isinstance(
        settings_2d.processing.color.balance, zivid.Settings2D.Processing.Color.Balance
    )
    assert settings_2d.processing.color.gamma is None
    assert settings_2d.processing.color.balance.red is None
    assert settings_2d.processing.color.balance.green is None
    assert settings_2d.processing.color.balance.blue is None


def test_set_acquisition_list(application):
    import zivid

    settings = zivid.Settings2D()

    settings.acquisitions = [
        zivid.Settings2D.Acquisition(gain=1.0),
        zivid.Settings2D.Acquisition(gain=2.0),
        zivid.Settings2D.Acquisition(gain=3.0),
    ]
    assert len(settings.acquisitions) == 3
    assert settings.acquisitions is not None
    assert isinstance(settings.acquisitions, list)
    for element in settings.acquisitions:
        assert isinstance(element, zivid.Settings2D.Acquisition)

    assert settings.acquisitions[0].gain == 1.0
    assert settings.acquisitions[1].gain == 2.0
    assert settings.acquisitions[2].gain == 3.0

    settings.acquisitions[0].gain = 4.0
    assert settings.acquisitions[0].gain == 4.0


def test_append_acquisition_list(application):
    import zivid

    settings = zivid.Settings2D()
    settings.acquisitions.append(zivid.Settings2D.Acquisition(brightness=1.1))
    assert len(settings.acquisitions) == 1
    assert settings.acquisitions[0].brightness == 1.1


def test_default_acquisition(application):
    import zivid
    import datetime

    settings_2d = zivid.Settings2D(acquisitions=[zivid.Settings2D.Acquisition()])
    assert isinstance(settings_2d.acquisitions, list)
    acquisition = settings_2d.acquisitions[0]

    assert isinstance(acquisition, zivid.Settings2D.Acquisition)
    assert acquisition.aperture is None
    assert acquisition.brightness is None
    assert acquisition.exposure_time is None
    assert acquisition.gain is None

    pytest.helpers.equality_tester(
        zivid.Settings2D.Acquisition,
        [5, 0.5, datetime.timedelta(microseconds=11000), 14],
        [5, 0.5, datetime.timedelta(microseconds=11001), 14],
    )


def test_acquisition_brightness(application):
    import numbers
    import zivid

    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings2D.Acquisition(),
        member="brightness",
        value=0.5,
        expected_data_type=numbers.Real,
    )


def test_acquisition_exposure_time(application):
    import datetime
    import zivid

    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings2D.Acquisition(),
        member="exposure_time",
        value=datetime.timedelta(microseconds=100000),
        expected_data_type=datetime.timedelta,
    )


def test_acquisition_gain(application):
    import numbers
    import zivid

    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings2D.Acquisition(),
        member="gain",
        value=14,
        expected_data_type=numbers.Real,
    )


def test_acquisition_aperture(application):
    import numbers
    import zivid

    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings2D.Acquisition(),
        member="aperture",
        value=20.5,
        expected_data_type=numbers.Real,
    )


def test_settings_processing(application):
    import zivid

    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings2D(),
        member="processing",
        value=zivid.Settings2D.Processing(),
        expected_data_type=zivid.Settings2D.Processing,
    )
    pytest.helpers.equality_tester(
        zivid.Settings2D.Processing,
        [
            zivid.Settings2D.Processing.Color(
                0.9, zivid.Settings2D.Processing.Color.Balance(blue=1.1)
            )
        ],
        [
            zivid.Settings2D.Processing.Color(
                1.1, zivid.Settings2D.Processing.Color.Balance(blue=1.2)
            )
        ],
    )


def test_settings_processing_color(application):
    import zivid

    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings2D.Processing(),
        member="color",
        value=zivid.Settings2D.Processing.Color(),
        expected_data_type=zivid.Settings2D.Processing.Color,
    )
    pytest.helpers.equality_tester(
        zivid.Settings2D.Processing.Color,
        [0.9, zivid.Settings2D.Processing.Color.Balance(red=1.1)],
        [1.1, zivid.Settings2D.Processing.Color.Balance(red=1.2)],
    )


def test_settings_processing_color_gamma(
    application,
):
    import zivid
    import numbers

    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings2D.Processing.Color(),
        member="gamma",
        value=0.85,
        expected_data_type=numbers.Real,
    )


def test_settings_processing_color_balance(
    application,
):
    import zivid

    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings2D.Processing.Color(),
        member="balance",
        value=zivid.Settings2D.Processing.Color.Balance(),
        expected_data_type=zivid.Settings2D.Processing.Color.Balance,
    )
    pytest.helpers.equality_tester(
        zivid.Settings2D.Processing.Color.Balance,
        [1.1, 1.1, 1.1],
        [1.2, 1.1, 1.1],
    )


def test_settings_processing_color_balance_red(
    application,
):
    import zivid
    import numbers

    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings2D.Processing.Color.Balance(),
        member="red",
        value=2,
        expected_data_type=numbers.Real,
    )


def test_settings_processing_color_balance_green(
    application,
):
    import zivid
    import numbers

    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings2D.Processing.Color.Balance(),
        member="green",
        value=1.5,
        expected_data_type=numbers.Real,
    )


def test_settings_processing_color_balance_blue(
    application,
):
    import zivid
    import numbers

    pytest.helpers.set_attribute_tester(
        settings_instance=zivid.Settings2D.Processing.Color.Balance(),
        member="blue",
        value=1,
        expected_data_type=numbers.Real,
    )



import pytest


@pytest.mark.physical_camera
def test_write_user_data(physical_camera):
    physical_camera.write_user_data(b"This is my data")


def test_write_invalid_user_data(file_camera):
    with pytest.raises(TypeError):
        file_camera.write_user_data("This is my data")
    with pytest.raises(TypeError):
        file_camera.write_user_data(1)
    with pytest.raises(TypeError):
        file_camera.write_user_data([1, 2, 3])


@pytest.mark.physical_camera
def test_read_user_data(physical_camera):
    assert isinstance(physical_camera.user_data, bytes)



"""Contains a the Image class."""

import numpy

import _zivid


class Image:
    """A two-dimensional image stored on the host."""

    def __init__(self, impl):
        """Initialize Image wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        allowed_types = (
            _zivid.ImageRGBA,
            _zivid.ImageBGRA,
            _zivid.ImageSRGB,
        )
        if not isinstance(impl, allowed_types):
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected one of {}".format(
                    type(impl), ", ".join(allowed_types)
                ),
            )
        self.__impl = impl

    @property
    def height(self):
        """Get the height of the image (number of rows).

        Returns:
            A positive integer
        """
        return self.__impl.height()

    @property
    def width(self):
        """Get the width of the image (number of columns).

        Returns:
            A positive integer
        """
        return self.__impl.width()

    def save(self, file_path):
        """Save the image to a file.

        The supported file type is PNG with extension .png.
        This method will throw an exception if failing to save to the provided file_path.

        Args:
            file_path: A pathlib.Path instance or a string specifying destination
        """
        self.__impl.save(str(file_path))

    @classmethod
    def load(cls, file_path, color_format):
        r"""Load an image from a file.

        The supported file types are PNG (.png), JPEG (.jpg, .jpeg), and BMP (.bmp). This method
        will throw an exception if it fails to load the provided file_path.

        Supported color formats:
        rgba:       ndarray(Height,Width,4) of uint8
        bgra:       ndarray(Height,Width,4) of uint8
        srgb:       ndarray(Height,Width,4) of uint8

        Args:
            file_path: A pathlib.Path instance or a string specifying file path to load
            color_format: A string specifying color format to load

        Returns:
            A Zivid.Image object with requested color format

        Raises:
            ValueError: If the requested color format does not exist
        """
        supported_color_formats = {
            "rgba": _zivid.ImageRGBA,
            "bgra": _zivid.ImageBGRA,
            "srgb": _zivid.ImageSRGB,
        }
        if color_format not in supported_color_formats:
            raise ValueError(
                "Unsupported color format: {color_format}. Supported formats: {all_formats}".format(
                    color_format=color_format,
                    all_formats=list(supported_color_formats.keys()),
                )
            )
        color_format_class = supported_color_formats[color_format]
        return Image(color_format_class(str(file_path)))

    def copy_data(self):
        """Copy image data to numpy array.

        Returns:
            A numpy array containing color pixel data
        """
        self.__impl.assert_not_released()
        return numpy.array(self.__impl)

    def release(self):
        """Release the underlying resources."""
        try:
            impl = self.__impl
        except AttributeError:
            pass
        else:
            impl.release()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.release()

    def __del__(self):
        self.release()



def test_import_zivid_globals_changes():
    before = sorted(globals().keys())
    import zivid  # pylint: disable=unused-import

    after = sorted(globals().keys())
    assert before == after


def test_import_zivid_globals_locals():
    expected_changes = sorted(["before", "zivid"])
    before = sorted(locals().keys())
    import zivid  # pylint: disable=possibly-unused-variable

    after = sorted(locals().keys())
    assert sorted(before + expected_changes) == after


def test_version():
    import zivid
    import _zivid

    assert zivid.__version__
    assert _zivid.__version__
    assert zivid.__version__ == _zivid.__version__



"""Auto generated, do not edit."""

# pylint: disable=too-many-lines,protected-access,too-few-public-methods,too-many-arguments,line-too-long,missing-function-docstring,missing-class-docstring,redefined-builtin,too-many-branches,too-many-boolean-expressions
import _zivid


class CameraIntrinsics:

    class CameraMatrix:

        def __init__(
            self,
            cx=_zivid.CameraIntrinsics.CameraMatrix.CX().value,
            cy=_zivid.CameraIntrinsics.CameraMatrix.CY().value,
            fx=_zivid.CameraIntrinsics.CameraMatrix.FX().value,
            fy=_zivid.CameraIntrinsics.CameraMatrix.FY().value,
        ):

            if isinstance(
                cx,
                (
                    float,
                    int,
                ),
            ):
                self._cx = _zivid.CameraIntrinsics.CameraMatrix.CX(cx)
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,), got {value_type}".format(
                        value_type=type(cx)
                    )
                )

            if isinstance(
                cy,
                (
                    float,
                    int,
                ),
            ):
                self._cy = _zivid.CameraIntrinsics.CameraMatrix.CY(cy)
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,), got {value_type}".format(
                        value_type=type(cy)
                    )
                )

            if isinstance(
                fx,
                (
                    float,
                    int,
                ),
            ):
                self._fx = _zivid.CameraIntrinsics.CameraMatrix.FX(fx)
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,), got {value_type}".format(
                        value_type=type(fx)
                    )
                )

            if isinstance(
                fy,
                (
                    float,
                    int,
                ),
            ):
                self._fy = _zivid.CameraIntrinsics.CameraMatrix.FY(fy)
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,), got {value_type}".format(
                        value_type=type(fy)
                    )
                )

        @property
        def cx(self):
            return self._cx.value

        @property
        def cy(self):
            return self._cy.value

        @property
        def fx(self):
            return self._fx.value

        @property
        def fy(self):
            return self._fy.value

        @cx.setter
        def cx(self, value):
            if isinstance(
                value,
                (
                    float,
                    int,
                ),
            ):
                self._cx = _zivid.CameraIntrinsics.CameraMatrix.CX(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        @cy.setter
        def cy(self, value):
            if isinstance(
                value,
                (
                    float,
                    int,
                ),
            ):
                self._cy = _zivid.CameraIntrinsics.CameraMatrix.CY(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        @fx.setter
        def fx(self, value):
            if isinstance(
                value,
                (
                    float,
                    int,
                ),
            ):
                self._fx = _zivid.CameraIntrinsics.CameraMatrix.FX(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        @fy.setter
        def fy(self, value):
            if isinstance(
                value,
                (
                    float,
                    int,
                ),
            ):
                self._fy = _zivid.CameraIntrinsics.CameraMatrix.FY(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        def __eq__(self, other):
            if (
                self._cx == other._cx
                and self._cy == other._cy
                and self._fx == other._fx
                and self._fy == other._fy
            ):
                return True
            return False

        def __str__(self):
            return str(_to_internal_camera_intrinsics_camera_matrix(self))

    class Distortion:

        def __init__(
            self,
            k1=_zivid.CameraIntrinsics.Distortion.K1().value,
            k2=_zivid.CameraIntrinsics.Distortion.K2().value,
            k3=_zivid.CameraIntrinsics.Distortion.K3().value,
            p1=_zivid.CameraIntrinsics.Distortion.P1().value,
            p2=_zivid.CameraIntrinsics.Distortion.P2().value,
        ):

            if isinstance(
                k1,
                (
                    float,
                    int,
                ),
            ):
                self._k1 = _zivid.CameraIntrinsics.Distortion.K1(k1)
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,), got {value_type}".format(
                        value_type=type(k1)
                    )
                )

            if isinstance(
                k2,
                (
                    float,
                    int,
                ),
            ):
                self._k2 = _zivid.CameraIntrinsics.Distortion.K2(k2)
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,), got {value_type}".format(
                        value_type=type(k2)
                    )
                )

            if isinstance(
                k3,
                (
                    float,
                    int,
                ),
            ):
                self._k3 = _zivid.CameraIntrinsics.Distortion.K3(k3)
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,), got {value_type}".format(
                        value_type=type(k3)
                    )
                )

            if isinstance(
                p1,
                (
                    float,
                    int,
                ),
            ):
                self._p1 = _zivid.CameraIntrinsics.Distortion.P1(p1)
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,), got {value_type}".format(
                        value_type=type(p1)
                    )
                )

            if isinstance(
                p2,
                (
                    float,
                    int,
                ),
            ):
                self._p2 = _zivid.CameraIntrinsics.Distortion.P2(p2)
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,), got {value_type}".format(
                        value_type=type(p2)
                    )
                )

        @property
        def k1(self):
            return self._k1.value

        @property
        def k2(self):
            return self._k2.value

        @property
        def k3(self):
            return self._k3.value

        @property
        def p1(self):
            return self._p1.value

        @property
        def p2(self):
            return self._p2.value

        @k1.setter
        def k1(self, value):
            if isinstance(
                value,
                (
                    float,
                    int,
                ),
            ):
                self._k1 = _zivid.CameraIntrinsics.Distortion.K1(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        @k2.setter
        def k2(self, value):
            if isinstance(
                value,
                (
                    float,
                    int,
                ),
            ):
                self._k2 = _zivid.CameraIntrinsics.Distortion.K2(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        @k3.setter
        def k3(self, value):
            if isinstance(
                value,
                (
                    float,
                    int,
                ),
            ):
                self._k3 = _zivid.CameraIntrinsics.Distortion.K3(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        @p1.setter
        def p1(self, value):
            if isinstance(
                value,
                (
                    float,
                    int,
                ),
            ):
                self._p1 = _zivid.CameraIntrinsics.Distortion.P1(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        @p2.setter
        def p2(self, value):
            if isinstance(
                value,
                (
                    float,
                    int,
                ),
            ):
                self._p2 = _zivid.CameraIntrinsics.Distortion.P2(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        def __eq__(self, other):
            if (
                self._k1 == other._k1
                and self._k2 == other._k2
                and self._k3 == other._k3
                and self._p1 == other._p1
                and self._p2 == other._p2
            ):
                return True
            return False

        def __str__(self):
            return str(_to_internal_camera_intrinsics_distortion(self))

    def __init__(
        self,
        camera_matrix=None,
        distortion=None,
    ):

        if camera_matrix is None:
            camera_matrix = self.CameraMatrix()
        if not isinstance(camera_matrix, self.CameraMatrix):
            raise TypeError(
                "Unsupported type: {value}".format(value=type(camera_matrix))
            )
        self._camera_matrix = camera_matrix

        if distortion is None:
            distortion = self.Distortion()
        if not isinstance(distortion, self.Distortion):
            raise TypeError("Unsupported type: {value}".format(value=type(distortion)))
        self._distortion = distortion

    @property
    def camera_matrix(self):
        return self._camera_matrix

    @property
    def distortion(self):
        return self._distortion

    @camera_matrix.setter
    def camera_matrix(self, value):
        if not isinstance(value, self.CameraMatrix):
            raise TypeError("Unsupported type {value}".format(value=type(value)))
        self._camera_matrix = value

    @distortion.setter
    def distortion(self, value):
        if not isinstance(value, self.Distortion):
            raise TypeError("Unsupported type {value}".format(value=type(value)))
        self._distortion = value

    @classmethod
    def load(cls, file_name):
        return _to_camera_intrinsics(_zivid.CameraIntrinsics(str(file_name)))

    def save(self, file_name):
        _to_internal_camera_intrinsics(self).save(str(file_name))

    @classmethod
    def from_serialized(cls, value):
        return _to_camera_intrinsics(
            _zivid.CameraIntrinsics.from_serialized(str(value))
        )

    def serialize(self):
        return _to_internal_camera_intrinsics(self).serialize()

    def __eq__(self, other):
        if (
            self._camera_matrix == other._camera_matrix
            and self._distortion == other._distortion
        ):
            return True
        return False

    def __str__(self):
        return str(_to_internal_camera_intrinsics(self))


def _to_camera_intrinsics_camera_matrix(internal_camera_matrix):
    return CameraIntrinsics.CameraMatrix(
        cx=internal_camera_matrix.cx.value,
        cy=internal_camera_matrix.cy.value,
        fx=internal_camera_matrix.fx.value,
        fy=internal_camera_matrix.fy.value,
    )


def _to_camera_intrinsics_distortion(internal_distortion):
    return CameraIntrinsics.Distortion(
        k1=internal_distortion.k1.value,
        k2=internal_distortion.k2.value,
        k3=internal_distortion.k3.value,
        p1=internal_distortion.p1.value,
        p2=internal_distortion.p2.value,
    )


def _to_camera_intrinsics(internal_camera_intrinsics):
    return CameraIntrinsics(
        camera_matrix=_to_camera_intrinsics_camera_matrix(
            internal_camera_intrinsics.camera_matrix
        ),
        distortion=_to_camera_intrinsics_distortion(
            internal_camera_intrinsics.distortion
        ),
    )


def _to_internal_camera_intrinsics_camera_matrix(camera_matrix):
    internal_camera_matrix = _zivid.CameraIntrinsics.CameraMatrix()

    internal_camera_matrix.cx = _zivid.CameraIntrinsics.CameraMatrix.CX(
        camera_matrix.cx
    )
    internal_camera_matrix.cy = _zivid.CameraIntrinsics.CameraMatrix.CY(
        camera_matrix.cy
    )
    internal_camera_matrix.fx = _zivid.CameraIntrinsics.CameraMatrix.FX(
        camera_matrix.fx
    )
    internal_camera_matrix.fy = _zivid.CameraIntrinsics.CameraMatrix.FY(
        camera_matrix.fy
    )

    return internal_camera_matrix


def _to_internal_camera_intrinsics_distortion(distortion):
    internal_distortion = _zivid.CameraIntrinsics.Distortion()

    internal_distortion.k1 = _zivid.CameraIntrinsics.Distortion.K1(distortion.k1)
    internal_distortion.k2 = _zivid.CameraIntrinsics.Distortion.K2(distortion.k2)
    internal_distortion.k3 = _zivid.CameraIntrinsics.Distortion.K3(distortion.k3)
    internal_distortion.p1 = _zivid.CameraIntrinsics.Distortion.P1(distortion.p1)
    internal_distortion.p2 = _zivid.CameraIntrinsics.Distortion.P2(distortion.p2)

    return internal_distortion


def _to_internal_camera_intrinsics(camera_intrinsics):
    internal_camera_intrinsics = _zivid.CameraIntrinsics()

    internal_camera_intrinsics.camera_matrix = (
        _to_internal_camera_intrinsics_camera_matrix(camera_intrinsics.camera_matrix)
    )
    internal_camera_intrinsics.distortion = _to_internal_camera_intrinsics_distortion(
        camera_intrinsics.distortion
    )
    return internal_camera_intrinsics



import pytest
import zivid


class ScopeExit:
    def __init__(self, exit_func):
        self.exit_func = exit_func

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.exit_func()


class DisconnectCamera(ScopeExit):
    def __init__(self, camera):
        camera.disconnect()
        super().__init__(lambda: camera.connect)


class RestoreNetworkConfiguration(ScopeExit):
    def __init__(self, camera):
        network_configuration = camera.network_configuration
        super().__init__(
            lambda: camera.apply_network_configuration(network_configuration)
        )


def test_default_network_configuration():
    network_configuration = zivid.NetworkConfiguration()
    assert (
        network_configuration.ipv4.mode == zivid.NetworkConfiguration.IPV4.Mode.manual
    )
    assert network_configuration.ipv4.address == "172.28.60.5"
    assert network_configuration.ipv4.subnet_mask == "255.255.255.0"


@pytest.mark.physical_camera
def test_fetch_network_configuration_while_connected(physical_camera):
    network_configuration = physical_camera.network_configuration
    assert network_configuration is not None
    assert isinstance(network_configuration, zivid.NetworkConfiguration)
    assert (
        network_configuration.ipv4.mode
        in zivid.NetworkConfiguration.IPV4.Mode.valid_values()
    )
    assert network_configuration.ipv4.address
    assert network_configuration.ipv4.subnet_mask


@pytest.mark.physical_camera
def test_fetch_network_configuration_while_not_connected(physical_camera):
    with DisconnectCamera(physical_camera):
        network_configuration = physical_camera.network_configuration
        assert network_configuration is not None
        assert isinstance(network_configuration, zivid.NetworkConfiguration)
        assert (
            network_configuration.ipv4.mode
            in zivid.NetworkConfiguration.IPV4.Mode.valid_values()
        )
        assert network_configuration.ipv4.address
        assert network_configuration.ipv4.subnet_mask


@pytest.mark.physical_camera
def test_apply_network_configuration_fails_while_connected(physical_camera):
    with pytest.raises(RuntimeError):
        physical_camera.apply_network_configuration(zivid.NetworkConfiguration())


@pytest.mark.physical_camera
def test_apply_default_network_configuration(physical_camera):
    with DisconnectCamera(physical_camera):
        with RestoreNetworkConfiguration(physical_camera):
            physical_camera.apply_network_configuration(zivid.NetworkConfiguration())
            assert physical_camera.network_configuration == zivid.NetworkConfiguration()



import pytest


def test_capture_from_file(file_camera_file):
    pytest.helpers.run_sample(
        name="capture_from_file", working_directory=file_camera_file.parent
    )


def test_print_version_info():
    pytest.helpers.run_sample(name="print_version_info")


@pytest.mark.physical_camera
def test_capture():
    pytest.helpers.run_sample(name="capture")


@pytest.mark.physical_camera
def test_capture_2d():
    pytest.helpers.run_sample(name="capture_2d")


@pytest.mark.physical_camera
def test_capture_assistant():
    pytest.helpers.run_sample(name="capture_assistant")


@pytest.mark.physical_camera
def test_capture_hdr():
    pytest.helpers.run_sample(name="capture_hdr")



"""Auto generated, do not edit."""

# pylint: disable=too-many-lines,protected-access,too-few-public-methods,too-many-arguments,line-too-long,missing-function-docstring,missing-class-docstring,redefined-builtin,too-many-branches,too-many-boolean-expressions
import _zivid


class CameraInfo:

    class Revision:

        def __init__(
            self,
            major=_zivid.CameraInfo.Revision.Major().value,
            minor=_zivid.CameraInfo.Revision.Minor().value,
        ):

            if isinstance(major, (int,)):
                self._major = _zivid.CameraInfo.Revision.Major(major)
            else:
                raise TypeError(
                    "Unsupported type, expected: (int,), got {value_type}".format(
                        value_type=type(major)
                    )
                )

            if isinstance(minor, (int,)):
                self._minor = _zivid.CameraInfo.Revision.Minor(minor)
            else:
                raise TypeError(
                    "Unsupported type, expected: (int,), got {value_type}".format(
                        value_type=type(minor)
                    )
                )

        @property
        def major(self):
            return self._major.value

        @property
        def minor(self):
            return self._minor.value

        @major.setter
        def major(self, value):
            if isinstance(value, (int,)):
                self._major = _zivid.CameraInfo.Revision.Major(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: int, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        @minor.setter
        def minor(self, value):
            if isinstance(value, (int,)):
                self._minor = _zivid.CameraInfo.Revision.Minor(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: int, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        def __eq__(self, other):
            if self._major == other._major and self._minor == other._minor:
                return True
            return False

        def __str__(self):
            return str(_to_internal_camera_info_revision(self))

    class UserData:

        def __init__(
            self,
            max_size_bytes=_zivid.CameraInfo.UserData.MaxSizeBytes().value,
        ):

            if isinstance(max_size_bytes, (int,)):
                self._max_size_bytes = _zivid.CameraInfo.UserData.MaxSizeBytes(
                    max_size_bytes
                )
            else:
                raise TypeError(
                    "Unsupported type, expected: (int,), got {value_type}".format(
                        value_type=type(max_size_bytes)
                    )
                )

        @property
        def max_size_bytes(self):
            return self._max_size_bytes.value

        @max_size_bytes.setter
        def max_size_bytes(self, value):
            if isinstance(value, (int,)):
                self._max_size_bytes = _zivid.CameraInfo.UserData.MaxSizeBytes(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: int, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        def __eq__(self, other):
            if self._max_size_bytes == other._max_size_bytes:
                return True
            return False

        def __str__(self):
            return str(_to_internal_camera_info_user_data(self))

    class Model:

        zivid2PlusL110 = "zivid2PlusL110"
        zivid2PlusLR110 = "zivid2PlusLR110"
        zivid2PlusM130 = "zivid2PlusM130"
        zivid2PlusM60 = "zivid2PlusM60"
        zivid2PlusMR130 = "zivid2PlusMR130"
        zivid2PlusMR60 = "zivid2PlusMR60"
        zividOnePlusLarge = "zividOnePlusLarge"
        zividOnePlusMedium = "zividOnePlusMedium"
        zividOnePlusSmall = "zividOnePlusSmall"
        zividTwo = "zividTwo"
        zividTwoL100 = "zividTwoL100"

        _valid_values = {
            "zivid2PlusL110": _zivid.CameraInfo.Model.zivid2PlusL110,
            "zivid2PlusLR110": _zivid.CameraInfo.Model.zivid2PlusLR110,
            "zivid2PlusM130": _zivid.CameraInfo.Model.zivid2PlusM130,
            "zivid2PlusM60": _zivid.CameraInfo.Model.zivid2PlusM60,
            "zivid2PlusMR130": _zivid.CameraInfo.Model.zivid2PlusMR130,
            "zivid2PlusMR60": _zivid.CameraInfo.Model.zivid2PlusMR60,
            "zividOnePlusLarge": _zivid.CameraInfo.Model.zividOnePlusLarge,
            "zividOnePlusMedium": _zivid.CameraInfo.Model.zividOnePlusMedium,
            "zividOnePlusSmall": _zivid.CameraInfo.Model.zividOnePlusSmall,
            "zividTwo": _zivid.CameraInfo.Model.zividTwo,
            "zividTwoL100": _zivid.CameraInfo.Model.zividTwoL100,
        }

        @classmethod
        def valid_values(cls):
            return list(cls._valid_values.keys())

    def __init__(
        self,
        firmware_version=_zivid.CameraInfo.FirmwareVersion().value,
        hardware_revision=_zivid.CameraInfo.HardwareRevision().value,
        model=_zivid.CameraInfo.Model().value,
        model_name=_zivid.CameraInfo.ModelName().value,
        serial_number=_zivid.CameraInfo.SerialNumber().value,
        revision=None,
        user_data=None,
    ):

        if isinstance(firmware_version, (str,)):
            self._firmware_version = _zivid.CameraInfo.FirmwareVersion(firmware_version)
        else:
            raise TypeError(
                "Unsupported type, expected: (str,), got {value_type}".format(
                    value_type=type(firmware_version)
                )
            )

        if isinstance(hardware_revision, (str,)):
            self._hardware_revision = _zivid.CameraInfo.HardwareRevision(
                hardware_revision
            )
        else:
            raise TypeError(
                "Unsupported type, expected: (str,), got {value_type}".format(
                    value_type=type(hardware_revision)
                )
            )

        if isinstance(model, _zivid.CameraInfo.Model.enum):
            self._model = _zivid.CameraInfo.Model(model)
        elif isinstance(model, str):
            self._model = _zivid.CameraInfo.Model(self.Model._valid_values[model])
        else:
            raise TypeError(
                "Unsupported type, expected: str, got {value_type}".format(
                    value_type=type(model)
                )
            )

        if isinstance(model_name, (str,)):
            self._model_name = _zivid.CameraInfo.ModelName(model_name)
        else:
            raise TypeError(
                "Unsupported type, expected: (str,), got {value_type}".format(
                    value_type=type(model_name)
                )
            )

        if isinstance(serial_number, (str,)):
            self._serial_number = _zivid.CameraInfo.SerialNumber(serial_number)
        else:
            raise TypeError(
                "Unsupported type, expected: (str,), got {value_type}".format(
                    value_type=type(serial_number)
                )
            )

        if revision is None:
            revision = self.Revision()
        if not isinstance(revision, self.Revision):
            raise TypeError("Unsupported type: {value}".format(value=type(revision)))
        self._revision = revision

        if user_data is None:
            user_data = self.UserData()
        if not isinstance(user_data, self.UserData):
            raise TypeError("Unsupported type: {value}".format(value=type(user_data)))
        self._user_data = user_data

    @property
    def firmware_version(self):
        return self._firmware_version.value

    @property
    def hardware_revision(self):
        return self._hardware_revision.value

    @property
    def model(self):
        if self._model.value is None:
            return None
        for key, internal_value in self.Model._valid_values.items():
            if internal_value == self._model.value:
                return key
        raise ValueError("Unsupported value {value}".format(value=self._model))

    @property
    def model_name(self):
        return self._model_name.value

    @property
    def serial_number(self):
        return self._serial_number.value

    @property
    def revision(self):
        return self._revision

    @property
    def user_data(self):
        return self._user_data

    @firmware_version.setter
    def firmware_version(self, value):
        if isinstance(value, (str,)):
            self._firmware_version = _zivid.CameraInfo.FirmwareVersion(value)
        else:
            raise TypeError(
                "Unsupported type, expected: str, got {value_type}".format(
                    value_type=type(value)
                )
            )

    @hardware_revision.setter
    def hardware_revision(self, value):
        if isinstance(value, (str,)):
            self._hardware_revision = _zivid.CameraInfo.HardwareRevision(value)
        else:
            raise TypeError(
                "Unsupported type, expected: str, got {value_type}".format(
                    value_type=type(value)
                )
            )

    @model.setter
    def model(self, value):
        if isinstance(value, str):
            self._model = _zivid.CameraInfo.Model(self.Model._valid_values[value])
        elif isinstance(value, _zivid.CameraInfo.Model.enum):
            self._model = _zivid.CameraInfo.Model(value)
        else:
            raise TypeError(
                "Unsupported type, expected: str, got {value_type}".format(
                    value_type=type(value)
                )
            )

    @model_name.setter
    def model_name(self, value):
        if isinstance(value, (str,)):
            self._model_name = _zivid.CameraInfo.ModelName(value)
        else:
            raise TypeError(
                "Unsupported type, expected: str, got {value_type}".format(
                    value_type=type(value)
                )
            )

    @serial_number.setter
    def serial_number(self, value):
        if isinstance(value, (str,)):
            self._serial_number = _zivid.CameraInfo.SerialNumber(value)
        else:
            raise TypeError(
                "Unsupported type, expected: str, got {value_type}".format(
                    value_type=type(value)
                )
            )

    @revision.setter
    def revision(self, value):
        if not isinstance(value, self.Revision):
            raise TypeError("Unsupported type {value}".format(value=type(value)))
        self._revision = value

    @user_data.setter
    def user_data(self, value):
        if not isinstance(value, self.UserData):
            raise TypeError("Unsupported type {value}".format(value=type(value)))
        self._user_data = value

    @classmethod
    def load(cls, file_name):
        return _to_camera_info(_zivid.CameraInfo(str(file_name)))

    def save(self, file_name):
        _to_internal_camera_info(self).save(str(file_name))

    @classmethod
    def from_serialized(cls, value):
        return _to_camera_info(_zivid.CameraInfo.from_serialized(str(value)))

    def serialize(self):
        return _to_internal_camera_info(self).serialize()

    def __eq__(self, other):
        if (
            self._firmware_version == other._firmware_version
            and self._hardware_revision == other._hardware_revision
            and self._model == other._model
            and self._model_name == other._model_name
            and self._serial_number == other._serial_number
            and self._revision == other._revision
            and self._user_data == other._user_data
        ):
            return True
        return False

    def __str__(self):
        return str(_to_internal_camera_info(self))


def _to_camera_info_revision(internal_revision):
    return CameraInfo.Revision(
        major=internal_revision.major.value,
        minor=internal_revision.minor.value,
    )


def _to_camera_info_user_data(internal_user_data):
    return CameraInfo.UserData(
        max_size_bytes=internal_user_data.max_size_bytes.value,
    )


def _to_camera_info(internal_camera_info):
    return CameraInfo(
        revision=_to_camera_info_revision(internal_camera_info.revision),
        user_data=_to_camera_info_user_data(internal_camera_info.user_data),
        firmware_version=internal_camera_info.firmware_version.value,
        hardware_revision=internal_camera_info.hardware_revision.value,
        model=internal_camera_info.model.value,
        model_name=internal_camera_info.model_name.value,
        serial_number=internal_camera_info.serial_number.value,
    )


def _to_internal_camera_info_revision(revision):
    internal_revision = _zivid.CameraInfo.Revision()

    internal_revision.major = _zivid.CameraInfo.Revision.Major(revision.major)
    internal_revision.minor = _zivid.CameraInfo.Revision.Minor(revision.minor)

    return internal_revision


def _to_internal_camera_info_user_data(user_data):
    internal_user_data = _zivid.CameraInfo.UserData()

    internal_user_data.max_size_bytes = _zivid.CameraInfo.UserData.MaxSizeBytes(
        user_data.max_size_bytes
    )

    return internal_user_data


def _to_internal_camera_info(camera_info):
    internal_camera_info = _zivid.CameraInfo()

    internal_camera_info.firmware_version = _zivid.CameraInfo.FirmwareVersion(
        camera_info.firmware_version
    )
    internal_camera_info.hardware_revision = _zivid.CameraInfo.HardwareRevision(
        camera_info.hardware_revision
    )
    internal_camera_info.model = _zivid.CameraInfo.Model(camera_info._model.value)
    internal_camera_info.model_name = _zivid.CameraInfo.ModelName(
        camera_info.model_name
    )
    internal_camera_info.serial_number = _zivid.CameraInfo.SerialNumber(
        camera_info.serial_number
    )

    internal_camera_info.revision = _to_internal_camera_info_revision(
        camera_info.revision
    )
    internal_camera_info.user_data = _to_internal_camera_info_user_data(
        camera_info.user_data
    )
    return internal_camera_info



"""Auto generated, do not edit."""

# pylint: disable=too-many-lines,protected-access,too-few-public-methods,too-many-arguments,line-too-long,missing-function-docstring,missing-class-docstring,redefined-builtin,too-many-branches,too-many-boolean-expressions
import datetime
import collections.abc
import _zivid


class Settings2D:

    class Acquisition:

        def __init__(
            self,
            aperture=_zivid.Settings2D.Acquisition.Aperture().value,
            brightness=_zivid.Settings2D.Acquisition.Brightness().value,
            exposure_time=_zivid.Settings2D.Acquisition.ExposureTime().value,
            gain=_zivid.Settings2D.Acquisition.Gain().value,
        ):

            if (
                isinstance(
                    aperture,
                    (
                        float,
                        int,
                    ),
                )
                or aperture is None
            ):
                self._aperture = _zivid.Settings2D.Acquisition.Aperture(aperture)
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,) or None, got {value_type}".format(
                        value_type=type(aperture)
                    )
                )

            if (
                isinstance(
                    brightness,
                    (
                        float,
                        int,
                    ),
                )
                or brightness is None
            ):
                self._brightness = _zivid.Settings2D.Acquisition.Brightness(brightness)
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,) or None, got {value_type}".format(
                        value_type=type(brightness)
                    )
                )

            if (
                isinstance(exposure_time, (datetime.timedelta,))
                or exposure_time is None
            ):
                self._exposure_time = _zivid.Settings2D.Acquisition.ExposureTime(
                    exposure_time
                )
            else:
                raise TypeError(
                    "Unsupported type, expected: (datetime.timedelta,) or None, got {value_type}".format(
                        value_type=type(exposure_time)
                    )
                )

            if (
                isinstance(
                    gain,
                    (
                        float,
                        int,
                    ),
                )
                or gain is None
            ):
                self._gain = _zivid.Settings2D.Acquisition.Gain(gain)
            else:
                raise TypeError(
                    "Unsupported type, expected: (float, int,) or None, got {value_type}".format(
                        value_type=type(gain)
                    )
                )

        @property
        def aperture(self):
            return self._aperture.value

        @property
        def brightness(self):
            return self._brightness.value

        @property
        def exposure_time(self):
            return self._exposure_time.value

        @property
        def gain(self):
            return self._gain.value

        @aperture.setter
        def aperture(self, value):
            if (
                isinstance(
                    value,
                    (
                        float,
                        int,
                    ),
                )
                or value is None
            ):
                self._aperture = _zivid.Settings2D.Acquisition.Aperture(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int or None, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        @brightness.setter
        def brightness(self, value):
            if (
                isinstance(
                    value,
                    (
                        float,
                        int,
                    ),
                )
                or value is None
            ):
                self._brightness = _zivid.Settings2D.Acquisition.Brightness(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int or None, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        @exposure_time.setter
        def exposure_time(self, value):
            if isinstance(value, (datetime.timedelta,)) or value is None:
                self._exposure_time = _zivid.Settings2D.Acquisition.ExposureTime(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: datetime.timedelta or None, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        @gain.setter
        def gain(self, value):
            if (
                isinstance(
                    value,
                    (
                        float,
                        int,
                    ),
                )
                or value is None
            ):
                self._gain = _zivid.Settings2D.Acquisition.Gain(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: float or  int or None, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        def __eq__(self, other):
            if (
                self._aperture == other._aperture
                and self._brightness == other._brightness
                and self._exposure_time == other._exposure_time
                and self._gain == other._gain
            ):
                return True
            return False

        def __str__(self):
            return str(_to_internal_settings2d_acquisition(self))

    class Processing:

        class Color:

            class Balance:

                def __init__(
                    self,
                    blue=_zivid.Settings2D.Processing.Color.Balance.Blue().value,
                    green=_zivid.Settings2D.Processing.Color.Balance.Green().value,
                    red=_zivid.Settings2D.Processing.Color.Balance.Red().value,
                ):

                    if (
                        isinstance(
                            blue,
                            (
                                float,
                                int,
                            ),
                        )
                        or blue is None
                    ):
                        self._blue = _zivid.Settings2D.Processing.Color.Balance.Blue(
                            blue
                        )
                    else:
                        raise TypeError(
                            "Unsupported type, expected: (float, int,) or None, got {value_type}".format(
                                value_type=type(blue)
                            )
                        )

                    if (
                        isinstance(
                            green,
                            (
                                float,
                                int,
                            ),
                        )
                        or green is None
                    ):
                        self._green = _zivid.Settings2D.Processing.Color.Balance.Green(
                            green
                        )
                    else:
                        raise TypeError(
                            "Unsupported type, expected: (float, int,) or None, got {value_type}".format(
                                value_type=type(green)
                            )
                        )

                    if (
                        isinstance(
                            red,
                            (
                                float,
                                int,
                            ),
                        )
                        or red is None
                    ):
                        self._red = _zivid.Settings2D.Processing.Color.Balance.Red(red)
                    else:
                        raise TypeError(
                            "Unsupported type, expected: (float, int,) or None, got {value_type}".format(
                                value_type=type(red)
                            )
                        )

                @property
                def blue(self):
                    return self._blue.value

                @property
                def green(self):
                    return self._green.value

                @property
                def red(self):
                    return self._red.value

                @blue.setter
                def blue(self, value):
                    if (
                        isinstance(
                            value,
                            (
                                float,
                                int,
                            ),
                        )
                        or value is None
                    ):
                        self._blue = _zivid.Settings2D.Processing.Color.Balance.Blue(
                            value
                        )
                    else:
                        raise TypeError(
                            "Unsupported type, expected: float or  int or None, got {value_type}".format(
                                value_type=type(value)
                            )
                        )

                @green.setter
                def green(self, value):
                    if (
                        isinstance(
                            value,
                            (
                                float,
                                int,
                            ),
                        )
                        or value is None
                    ):
                        self._green = _zivid.Settings2D.Processing.Color.Balance.Green(
                            value
                        )
                    else:
                        raise TypeError(
                            "Unsupported type, expected: float or  int or None, got {value_type}".format(
                                value_type=type(value)
                            )
                        )

                @red.setter
                def red(self, value):
                    if (
                        isinstance(
                            value,
                            (
                                float,
                                int,
                            ),
                        )
                        or value is None
                    ):
                        self._red = _zivid.Settings2D.Processing.Color.Balance.Red(
                            value
                        )
                    else:
                        raise TypeError(
                            "Unsupported type, expected: float or  int or None, got {value_type}".format(
                                value_type=type(value)
                            )
                        )

                def __eq__(self, other):
                    if (
                        self._blue == other._blue
                        and self._green == other._green
                        and self._red == other._red
                    ):
                        return True
                    return False

                def __str__(self):
                    return str(_to_internal_settings2d_processing_color_balance(self))

            class Experimental:

                class Mode:

                    automatic = "automatic"
                    toneMapping = "toneMapping"

                    _valid_values = {
                        "automatic": _zivid.Settings2D.Processing.Color.Experimental.Mode.automatic,
                        "toneMapping": _zivid.Settings2D.Processing.Color.Experimental.Mode.toneMapping,
                    }

                    @classmethod
                    def valid_values(cls):
                        return list(cls._valid_values.keys())

                def __init__(
                    self,
                    mode=_zivid.Settings2D.Processing.Color.Experimental.Mode().value,
                ):

                    if (
                        isinstance(
                            mode,
                            _zivid.Settings2D.Processing.Color.Experimental.Mode.enum,
                        )
                        or mode is None
                    ):
                        self._mode = (
                            _zivid.Settings2D.Processing.Color.Experimental.Mode(mode)
                        )
                    elif isinstance(mode, str):
                        self._mode = (
                            _zivid.Settings2D.Processing.Color.Experimental.Mode(
                                self.Mode._valid_values[mode]
                            )
                        )
                    else:
                        raise TypeError(
                            "Unsupported type, expected: str or None, got {value_type}".format(
                                value_type=type(mode)
                            )
                        )

                @property
                def mode(self):
                    if self._mode.value is None:
                        return None
                    for key, internal_value in self.Mode._valid_values.items():
                        if internal_value == self._mode.value:
                            return key
                    raise ValueError(
                        "Unsupported value {value}".format(value=self._mode)
                    )

                @mode.setter
                def mode(self, value):
                    if isinstance(value, str):
                        self._mode = (
                            _zivid.Settings2D.Processing.Color.Experimental.Mode(
                                self.Mode._valid_values[value]
                            )
                        )
                    elif (
                        isinstance(
                            value,
                            _zivid.Settings2D.Processing.Color.Experimental.Mode.enum,
                        )
                        or value is None
                    ):
                        self._mode = (
                            _zivid.Settings2D.Processing.Color.Experimental.Mode(value)
                        )
                    else:
                        raise TypeError(
                            "Unsupported type, expected: str or None, got {value_type}".format(
                                value_type=type(value)
                            )
                        )

                def __eq__(self, other):
                    if self._mode == other._mode:
                        return True
                    return False

                def __str__(self):
                    return str(
                        _to_internal_settings2d_processing_color_experimental(self)
                    )

            def __init__(
                self,
                gamma=_zivid.Settings2D.Processing.Color.Gamma().value,
                balance=None,
                experimental=None,
            ):

                if (
                    isinstance(
                        gamma,
                        (
                            float,
                            int,
                        ),
                    )
                    or gamma is None
                ):
                    self._gamma = _zivid.Settings2D.Processing.Color.Gamma(gamma)
                else:
                    raise TypeError(
                        "Unsupported type, expected: (float, int,) or None, got {value_type}".format(
                            value_type=type(gamma)
                        )
                    )

                if balance is None:
                    balance = self.Balance()
                if not isinstance(balance, self.Balance):
                    raise TypeError(
                        "Unsupported type: {value}".format(value=type(balance))
                    )
                self._balance = balance

                if experimental is None:
                    experimental = self.Experimental()
                if not isinstance(experimental, self.Experimental):
                    raise TypeError(
                        "Unsupported type: {value}".format(value=type(experimental))
                    )
                self._experimental = experimental

            @property
            def gamma(self):
                return self._gamma.value

            @property
            def balance(self):
                return self._balance

            @property
            def experimental(self):
                return self._experimental

            @gamma.setter
            def gamma(self, value):
                if (
                    isinstance(
                        value,
                        (
                            float,
                            int,
                        ),
                    )
                    or value is None
                ):
                    self._gamma = _zivid.Settings2D.Processing.Color.Gamma(value)
                else:
                    raise TypeError(
                        "Unsupported type, expected: float or  int or None, got {value_type}".format(
                            value_type=type(value)
                        )
                    )

            @balance.setter
            def balance(self, value):
                if not isinstance(value, self.Balance):
                    raise TypeError(
                        "Unsupported type {value}".format(value=type(value))
                    )
                self._balance = value

            @experimental.setter
            def experimental(self, value):
                if not isinstance(value, self.Experimental):
                    raise TypeError(
                        "Unsupported type {value}".format(value=type(value))
                    )
                self._experimental = value

            def __eq__(self, other):
                if (
                    self._gamma == other._gamma
                    and self._balance == other._balance
                    and self._experimental == other._experimental
                ):
                    return True
                return False

            def __str__(self):
                return str(_to_internal_settings2d_processing_color(self))

        def __init__(
            self,
            color=None,
        ):

            if color is None:
                color = self.Color()
            if not isinstance(color, self.Color):
                raise TypeError("Unsupported type: {value}".format(value=type(color)))
            self._color = color

        @property
        def color(self):
            return self._color

        @color.setter
        def color(self, value):
            if not isinstance(value, self.Color):
                raise TypeError("Unsupported type {value}".format(value=type(value)))
            self._color = value

        def __eq__(self, other):
            if self._color == other._color:
                return True
            return False

        def __str__(self):
            return str(_to_internal_settings2d_processing(self))

    class Sampling:

        class Color:

            grayscale = "grayscale"
            rgb = "rgb"

            _valid_values = {
                "grayscale": _zivid.Settings2D.Sampling.Color.grayscale,
                "rgb": _zivid.Settings2D.Sampling.Color.rgb,
            }

            @classmethod
            def valid_values(cls):
                return list(cls._valid_values.keys())

        class Pixel:

            all = "all"
            blueSubsample2x2 = "blueSubsample2x2"
            blueSubsample4x4 = "blueSubsample4x4"
            by2x2 = "by2x2"
            by4x4 = "by4x4"
            redSubsample2x2 = "redSubsample2x2"
            redSubsample4x4 = "redSubsample4x4"

            _valid_values = {
                "all": _zivid.Settings2D.Sampling.Pixel.all,
                "blueSubsample2x2": _zivid.Settings2D.Sampling.Pixel.blueSubsample2x2,
                "blueSubsample4x4": _zivid.Settings2D.Sampling.Pixel.blueSubsample4x4,
                "by2x2": _zivid.Settings2D.Sampling.Pixel.by2x2,
                "by4x4": _zivid.Settings2D.Sampling.Pixel.by4x4,
                "redSubsample2x2": _zivid.Settings2D.Sampling.Pixel.redSubsample2x2,
                "redSubsample4x4": _zivid.Settings2D.Sampling.Pixel.redSubsample4x4,
            }

            @classmethod
            def valid_values(cls):
                return list(cls._valid_values.keys())

        def __init__(
            self,
            color=_zivid.Settings2D.Sampling.Color().value,
            pixel=_zivid.Settings2D.Sampling.Pixel().value,
        ):

            if (
                isinstance(color, _zivid.Settings2D.Sampling.Color.enum)
                or color is None
            ):
                self._color = _zivid.Settings2D.Sampling.Color(color)
            elif isinstance(color, str):
                self._color = _zivid.Settings2D.Sampling.Color(
                    self.Color._valid_values[color]
                )
            else:
                raise TypeError(
                    "Unsupported type, expected: str or None, got {value_type}".format(
                        value_type=type(color)
                    )
                )

            if (
                isinstance(pixel, _zivid.Settings2D.Sampling.Pixel.enum)
                or pixel is None
            ):
                self._pixel = _zivid.Settings2D.Sampling.Pixel(pixel)
            elif isinstance(pixel, str):
                self._pixel = _zivid.Settings2D.Sampling.Pixel(
                    self.Pixel._valid_values[pixel]
                )
            else:
                raise TypeError(
                    "Unsupported type, expected: str or None, got {value_type}".format(
                        value_type=type(pixel)
                    )
                )

        @property
        def color(self):
            if self._color.value is None:
                return None
            for key, internal_value in self.Color._valid_values.items():
                if internal_value == self._color.value:
                    return key
            raise ValueError("Unsupported value {value}".format(value=self._color))

        @property
        def pixel(self):
            if self._pixel.value is None:
                return None
            for key, internal_value in self.Pixel._valid_values.items():
                if internal_value == self._pixel.value:
                    return key
            raise ValueError("Unsupported value {value}".format(value=self._pixel))

        @color.setter
        def color(self, value):
            if isinstance(value, str):
                self._color = _zivid.Settings2D.Sampling.Color(
                    self.Color._valid_values[value]
                )
            elif (
                isinstance(value, _zivid.Settings2D.Sampling.Color.enum)
                or value is None
            ):
                self._color = _zivid.Settings2D.Sampling.Color(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: str or None, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        @pixel.setter
        def pixel(self, value):
            if isinstance(value, str):
                self._pixel = _zivid.Settings2D.Sampling.Pixel(
                    self.Pixel._valid_values[value]
                )
            elif (
                isinstance(value, _zivid.Settings2D.Sampling.Pixel.enum)
                or value is None
            ):
                self._pixel = _zivid.Settings2D.Sampling.Pixel(value)
            else:
                raise TypeError(
                    "Unsupported type, expected: str or None, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        def __eq__(self, other):
            if self._color == other._color and self._pixel == other._pixel:
                return True
            return False

        def __str__(self):
            return str(_to_internal_settings2d_sampling(self))

    def __init__(
        self,
        acquisitions=None,
        processing=None,
        sampling=None,
    ):

        if acquisitions is None:
            self._acquisitions = []
        elif isinstance(acquisitions, (collections.abc.Iterable,)):
            self._acquisitions = []
            for item in acquisitions:
                if isinstance(item, self.Acquisition):
                    self._acquisitions.append(item)
                else:
                    raise TypeError(
                        "Unsupported type {item_type}".format(item_type=type(item))
                    )
        else:
            raise TypeError(
                "Unsupported type, expected: (collections.abc.Iterable,) or None, got {value_type}".format(
                    value_type=type(acquisitions)
                )
            )

        if processing is None:
            processing = self.Processing()
        if not isinstance(processing, self.Processing):
            raise TypeError("Unsupported type: {value}".format(value=type(processing)))
        self._processing = processing

        if sampling is None:
            sampling = self.Sampling()
        if not isinstance(sampling, self.Sampling):
            raise TypeError("Unsupported type: {value}".format(value=type(sampling)))
        self._sampling = sampling

    @property
    def acquisitions(self):
        return self._acquisitions

    @property
    def processing(self):
        return self._processing

    @property
    def sampling(self):
        return self._sampling

    @acquisitions.setter
    def acquisitions(self, value):
        if not isinstance(value, (collections.abc.Iterable,)):
            raise TypeError("Unsupported type {value}".format(value=type(value)))
        self._acquisitions = []
        for item in value:
            if isinstance(item, self.Acquisition):
                self._acquisitions.append(item)
            else:
                raise TypeError(
                    "Unsupported type {item_type}".format(item_type=type(item))
                )

    @processing.setter
    def processing(self, value):
        if not isinstance(value, self.Processing):
            raise TypeError("Unsupported type {value}".format(value=type(value)))
        self._processing = value

    @sampling.setter
    def sampling(self, value):
        if not isinstance(value, self.Sampling):
            raise TypeError("Unsupported type {value}".format(value=type(value)))
        self._sampling = value

    @classmethod
    def load(cls, file_name):
        return _to_settings2d(_zivid.Settings2D(str(file_name)))

    def save(self, file_name):
        _to_internal_settings2d(self).save(str(file_name))

    @classmethod
    def from_serialized(cls, value):
        return _to_settings2d(_zivid.Settings2D.from_serialized(str(value)))

    def serialize(self):
        return _to_internal_settings2d(self).serialize()

    def __eq__(self, other):
        if (
            self._acquisitions == other._acquisitions
            and self._processing == other._processing
            and self._sampling == other._sampling
        ):
            return True
        return False

    def __str__(self):
        return str(_to_internal_settings2d(self))


def _to_settings2d_acquisition(internal_acquisition):
    return Settings2D.Acquisition(
        aperture=internal_acquisition.aperture.value,
        brightness=internal_acquisition.brightness.value,
        exposure_time=internal_acquisition.exposure_time.value,
        gain=internal_acquisition.gain.value,
    )


def _to_settings2d_processing_color_balance(internal_balance):
    return Settings2D.Processing.Color.Balance(
        blue=internal_balance.blue.value,
        green=internal_balance.green.value,
        red=internal_balance.red.value,
    )


def _to_settings2d_processing_color_experimental(internal_experimental):
    return Settings2D.Processing.Color.Experimental(
        mode=internal_experimental.mode.value,
    )


def _to_settings2d_processing_color(internal_color):
    return Settings2D.Processing.Color(
        balance=_to_settings2d_processing_color_balance(internal_color.balance),
        experimental=_to_settings2d_processing_color_experimental(
            internal_color.experimental
        ),
        gamma=internal_color.gamma.value,
    )


def _to_settings2d_processing(internal_processing):
    return Settings2D.Processing(
        color=_to_settings2d_processing_color(internal_processing.color),
    )


def _to_settings2d_sampling(internal_sampling):
    return Settings2D.Sampling(
        color=internal_sampling.color.value,
        pixel=internal_sampling.pixel.value,
    )


def _to_settings2d(internal_settings2d):
    return Settings2D(
        acquisitions=[
            _to_settings2d_acquisition(value)
            for value in internal_settings2d.acquisitions.value
        ],
        processing=_to_settings2d_processing(internal_settings2d.processing),
        sampling=_to_settings2d_sampling(internal_settings2d.sampling),
    )


def _to_internal_settings2d_acquisition(acquisition):
    internal_acquisition = _zivid.Settings2D.Acquisition()

    internal_acquisition.aperture = _zivid.Settings2D.Acquisition.Aperture(
        acquisition.aperture
    )
    internal_acquisition.brightness = _zivid.Settings2D.Acquisition.Brightness(
        acquisition.brightness
    )
    internal_acquisition.exposure_time = _zivid.Settings2D.Acquisition.ExposureTime(
        acquisition.exposure_time
    )
    internal_acquisition.gain = _zivid.Settings2D.Acquisition.Gain(acquisition.gain)

    return internal_acquisition


def _to_internal_settings2d_processing_color_balance(balance):
    internal_balance = _zivid.Settings2D.Processing.Color.Balance()

    internal_balance.blue = _zivid.Settings2D.Processing.Color.Balance.Blue(
        balance.blue
    )
    internal_balance.green = _zivid.Settings2D.Processing.Color.Balance.Green(
        balance.green
    )
    internal_balance.red = _zivid.Settings2D.Processing.Color.Balance.Red(balance.red)

    return internal_balance


def _to_internal_settings2d_processing_color_experimental(experimental):
    internal_experimental = _zivid.Settings2D.Processing.Color.Experimental()

    internal_experimental.mode = _zivid.Settings2D.Processing.Color.Experimental.Mode(
        experimental._mode.value
    )

    return internal_experimental


def _to_internal_settings2d_processing_color(color):
    internal_color = _zivid.Settings2D.Processing.Color()

    internal_color.gamma = _zivid.Settings2D.Processing.Color.Gamma(color.gamma)

    internal_color.balance = _to_internal_settings2d_processing_color_balance(
        color.balance
    )
    internal_color.experimental = _to_internal_settings2d_processing_color_experimental(
        color.experimental
    )
    return internal_color


def _to_internal_settings2d_processing(processing):
    internal_processing = _zivid.Settings2D.Processing()

    internal_processing.color = _to_internal_settings2d_processing_color(
        processing.color
    )
    return internal_processing


def _to_internal_settings2d_sampling(sampling):
    internal_sampling = _zivid.Settings2D.Sampling()

    internal_sampling.color = _zivid.Settings2D.Sampling.Color(sampling._color.value)
    internal_sampling.pixel = _zivid.Settings2D.Sampling.Pixel(sampling._pixel.value)

    return internal_sampling


def _to_internal_settings2d(settings2d):
    internal_settings2d = _zivid.Settings2D()

    temp_acquisitions = _zivid.Settings2D.Acquisitions()
    for value in settings2d.acquisitions:
        temp_acquisitions.append(_to_internal_settings2d_acquisition(value))
    internal_settings2d.acquisitions = temp_acquisitions

    internal_settings2d.processing = _to_internal_settings2d_processing(
        settings2d.processing
    )
    internal_settings2d.sampling = _to_internal_settings2d_sampling(settings2d.sampling)
    return internal_settings2d



import pytest


def test_ambient_light_frequency():
    from zivid.capture_assistant import SuggestSettingsParameters

    assert str(SuggestSettingsParameters.AmbientLightFrequency.hz50) == "hz50"
    assert str(SuggestSettingsParameters.AmbientLightFrequency.hz60) == "hz60"
    assert str(SuggestSettingsParameters.AmbientLightFrequency.none) == "none"

    assert SuggestSettingsParameters.AmbientLightFrequency.hz50 == "hz50"
    assert SuggestSettingsParameters.AmbientLightFrequency.hz60 == "hz60"
    assert SuggestSettingsParameters.AmbientLightFrequency.none == "none"


def test_suggest_settings_parameters():
    import datetime
    from zivid.capture_assistant import SuggestSettingsParameters

    # Use constructor
    suggest_settings_parameters = SuggestSettingsParameters(
        max_capture_time=datetime.timedelta(milliseconds=1200),
        ambient_light_frequency=SuggestSettingsParameters.AmbientLightFrequency.hz50,
    )
    assert isinstance(suggest_settings_parameters.max_capture_time, datetime.timedelta)
    assert suggest_settings_parameters.max_capture_time == datetime.timedelta(
        seconds=1.2
    )
    assert isinstance(suggest_settings_parameters.ambient_light_frequency, str)
    assert (
        suggest_settings_parameters.ambient_light_frequency
        == SuggestSettingsParameters.AmbientLightFrequency.hz50
    )

    # Use setters
    new_time = datetime.timedelta(milliseconds=1800)
    new_freq = "hz60"
    suggest_settings_parameters.max_capture_time = new_time
    assert isinstance(suggest_settings_parameters.max_capture_time, datetime.timedelta)
    assert suggest_settings_parameters.max_capture_time == new_time
    suggest_settings_parameters.ambient_light_frequency = new_freq
    assert isinstance(suggest_settings_parameters.ambient_light_frequency, str)
    assert suggest_settings_parameters.ambient_light_frequency == new_freq


def test_suggest_settings_throws_if_budget_outside_range():
    import datetime
    from zivid.capture_assistant import SuggestSettingsParameters

    # too small
    with pytest.raises(IndexError):
        SuggestSettingsParameters(
            max_capture_time=datetime.timedelta(milliseconds=100),
            ambient_light_frequency=SuggestSettingsParameters.AmbientLightFrequency.hz50,
        )
    # too big
    with pytest.raises(IndexError):
        SuggestSettingsParameters(
            max_capture_time=datetime.timedelta(milliseconds=60000)
        )


def test_init_max_capture_time():
    import datetime
    from zivid.capture_assistant import SuggestSettingsParameters

    suggested_settings = SuggestSettingsParameters(
        max_capture_time=datetime.timedelta(milliseconds=1000)
    )
    max_capture_time = suggested_settings.max_capture_time
    assert max_capture_time is not None
    assert isinstance(max_capture_time, datetime.timedelta)
    assert max_capture_time == datetime.timedelta(milliseconds=1000)


def test_default_ambient_light_frequency():
    import datetime
    from zivid.capture_assistant import SuggestSettingsParameters

    suggested_settings = SuggestSettingsParameters(
        max_capture_time=datetime.timedelta(milliseconds=250)
    )
    ambient_light_frequency = suggested_settings.ambient_light_frequency
    assert ambient_light_frequency is not None
    assert isinstance(ambient_light_frequency, str)
    assert (
        ambient_light_frequency == SuggestSettingsParameters.AmbientLightFrequency.none
    )


def test_set_ambient_light_frequency():
    from zivid.capture_assistant import SuggestSettingsParameters

    suggested_settings = SuggestSettingsParameters()

    suggested_settings.ambient_light_frequency = (
        SuggestSettingsParameters.AmbientLightFrequency.hz50
    )
    assert (
        suggested_settings.ambient_light_frequency
        == SuggestSettingsParameters.AmbientLightFrequency.hz50
    )
    assert suggested_settings.ambient_light_frequency == "hz50"
    assert isinstance(suggested_settings.ambient_light_frequency, str)

    suggested_settings.ambient_light_frequency = (
        SuggestSettingsParameters.AmbientLightFrequency.hz60
    )
    assert (
        suggested_settings.ambient_light_frequency
        == SuggestSettingsParameters.AmbientLightFrequency.hz60
    )
    assert suggested_settings.ambient_light_frequency == "hz60"
    assert isinstance(suggested_settings.ambient_light_frequency, str)

    suggested_settings.ambient_light_frequency = (
        SuggestSettingsParameters.AmbientLightFrequency.none
    )
    assert (
        suggested_settings.ambient_light_frequency
        == SuggestSettingsParameters.AmbientLightFrequency.none
    )
    assert suggested_settings.ambient_light_frequency == "none"
    assert isinstance(suggested_settings.ambient_light_frequency, str)

    # AmbientLightFrequency enum is not optional, so should throw if given None
    with pytest.raises(TypeError):
        suggested_settings.ambient_light_frequency = None
    with pytest.raises(TypeError):
        SuggestSettingsParameters(ambient_light_frequency=None)


def test_suggest_settings_str():
    from zivid.capture_assistant import SuggestSettingsParameters

    string = str(SuggestSettingsParameters())
    assert string is not None
    assert isinstance(string, str)


def test_ambient_light_frequency_str():
    from zivid.capture_assistant import SuggestSettingsParameters

    string = str(SuggestSettingsParameters().ambient_light_frequency)
    assert string is not None
    assert isinstance(string, str)



"""Module containing implementation of multi-camera calibration functionality.

This module should not be imported directly by end-user, but rather accessed through
the zivid.calibration module.
"""

import _zivid


class MultiCameraResidual:
    """Class representing the estimated errors of a multi-camera calibration."""

    def __init__(self, impl):
        """Initialize MultiCameraResidual wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if not isinstance(impl, _zivid.calibration.MultiCameraResidual):
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}".format(
                    type(impl), _zivid.calibration.MultiCameraResidual
                )
            )
        self.__impl = impl

    def translation(self):
        """Get the average overlap error.

        Returns:
            Average overlap error in millimeters
        """
        return self.__impl.translation()

    def __str__(self):
        return str(self.__impl)


class MultiCameraOutput:
    """Class representing the result of a multi-camera calibration process."""

    def __init__(self, impl):
        """Initialize MultiCameraOutput wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if not isinstance(impl, _zivid.calibration.MultiCameraOutput):
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}".format(
                    type(impl), _zivid.calibration.MultiCameraOutput
                )
            )
        self.__impl = impl

    def valid(self):
        """Check validity of MultiCameraOutput.

        Returns:
            True if MultiCameraOutput is valid
        """
        return self.__impl.valid()

    def __bool__(self):
        return bool(self.__impl)

    def transforms(self):
        """Get multi-camera calibration transforms.

        Returns:
            List of 4x4 arrays, one for each camera
        """
        return self.__impl.transforms()

    def residuals(self):
        """Get multi-camera calibration residuals.

        Returns:
            List of MultiCameraResidual instances, one for each camera
        """
        return [
            MultiCameraResidual(internal_residual)
            for internal_residual in self.__impl.residuals()
        ]

    def __str__(self):
        return str(self.__impl)


def calibrate_multi_camera(detection_results):
    """Perform multi-camera calibration.

    Args:
        detection_results: List of DetectionResult, one for each camera

    Returns:
        A MultiCameraOutput instance
    """
    return MultiCameraOutput(
        _zivid.calibration.calibrate_multi_camera(
            [
                detection_result._DetectionResult__impl  # pylint: disable=protected-access
                for detection_result in detection_results
            ]
        )
    )



from pathlib import Path
import tempfile
import numpy
import numpy.testing
import pytest
import zivid

ZIVID_MATRIX_SAVE_LOAD_TOLERANCE_DECIMAL = 5


def to_float32_1d(arr):
    for i, element in enumerate(arr):
        arr[i] = float(numpy.float32(element))
    return arr


def to_float32_2d(arr):
    for row in arr:
        to_float32_1d(row)
    return arr


def assert_all_equal_flat(matrix: zivid.Matrix4x4, arr2) -> None:
    assert isinstance(matrix, zivid.Matrix4x4)
    assert len(arr2) == 16
    for element1, element2 in zip(matrix, arr2):
        numpy.testing.assert_almost_equal(
            element1, element2, ZIVID_MATRIX_SAVE_LOAD_TOLERANCE_DECIMAL
        )


def assert_all_equal_2d(matrix: zivid.Matrix4x4, arr2) -> None:
    assert isinstance(matrix, zivid.Matrix4x4)
    assert not isinstance(arr2, zivid.Matrix4x4)
    assert len(arr2) == 4
    for row in range(4):
        assert len(arr2[row]) == 4
        for col in range(4):
            numpy.testing.assert_almost_equal(
                matrix[row, col],
                arr2[row][col],
                ZIVID_MATRIX_SAVE_LOAD_TOLERANCE_DECIMAL,
            )


def sample_2d_list() -> list:
    return to_float32_2d(
        [
            [-8.3, 6.75, -2, 5.7],
            [-0.24, 1.49, 3.5, -4.25],
            [52, 0.98, 970.6, 75000],
            [-64.3, 15.4, -84.4, -13.4],
        ]
    )


def sample_1d_list() -> list:
    return to_float32_1d(
        [
            3.8,
            0.64,
            -9.55,
            63226,
            -0.3445,
            9.5,
            4.004,
            0.115,
            9999,
            -34,
            14,
            66.5,
            87.4,
            1.3,
            36.2,
            93.12,
        ]
    )


def test_default_init():
    assert_all_equal_flat(zivid.Matrix4x4(), [0.0] * 16)


def test_flat_array_init():
    assert_all_equal_flat(zivid.Matrix4x4(sample_1d_list()), sample_1d_list())
    assert_all_equal_flat(
        zivid.Matrix4x4(numpy.array(sample_1d_list())), numpy.array(sample_1d_list())
    )

    with pytest.raises(TypeError):
        zivid.Matrix4x4(range(42))

    with pytest.raises(TypeError):
        zivid.Matrix4x4(range(0))


def test_4x4_array_init():
    assert_all_equal_2d(zivid.Matrix4x4(sample_2d_list()), sample_2d_list())
    assert_all_equal_2d(
        zivid.Matrix4x4(sample_2d_list()), numpy.array(sample_2d_list())
    )

    with pytest.raises(TypeError):
        zivid.Matrix4x4([range(4), range(4, 9), range(9, 13), range(13, 17)])

    with pytest.raises(TypeError):
        zivid.Matrix4x4([[], range(4, 8), range(8, 12), range(12, 16)])

    with pytest.raises(TypeError):
        zivid.Matrix4x4([[]] * 4)


def test_getitem():
    matrix = zivid.Matrix4x4(sample_2d_list())

    for i in range(4):
        for j in range(4):
            assert matrix[i, j] == sample_2d_list()[i][j]

    for i in range(-4, 0):
        for j in range(-4, 0):
            assert matrix[i, j] == sample_2d_list()[i][j]

    with pytest.raises(TypeError):
        assert matrix[0] == 0

    with pytest.raises(TypeError):
        assert matrix[0, 0, 0] == 0

    with pytest.raises(TypeError):
        assert matrix[1.4, 1.0] == 0

    with pytest.raises(TypeError):
        assert matrix["0", "0"] == 0

    with pytest.raises(IndexError):
        assert matrix[1000, 0] == 0

    with pytest.raises(IndexError):
        assert matrix[0, 1000] == 0

    with pytest.raises(IndexError):
        assert matrix[0, -1000] == 0


def test_setitem():
    matrix = zivid.Matrix4x4()

    for i in range(4):
        for j in range(4):
            matrix[i, j] = sample_2d_list()[i][j]
            assert matrix[i, j] == sample_2d_list()[i][j]

    assert_all_equal_2d(matrix, sample_2d_list())

    for i in range(-4, 0):
        for j in range(-4, 0):
            matrix[i, j] = sample_2d_list()[i][j]
            assert matrix[i, j] == sample_2d_list()[i][j]

    assert_all_equal_2d(matrix, sample_2d_list())

    with pytest.raises(TypeError):
        matrix[0] = 0

    with pytest.raises(TypeError):
        matrix[0, 0, 0] = 0

    with pytest.raises(TypeError):
        matrix[1.4, 1.0] = 0

    with pytest.raises(TypeError):
        matrix["0", "0"] = 0

    with pytest.raises(TypeError):
        matrix[0, 0] = "42"

    with pytest.raises(TypeError):
        matrix[0, 0] = 10**1000

    with pytest.raises(IndexError):
        matrix[1000, 0] = 0

    with pytest.raises(IndexError):
        matrix[0, 1000] = 0

    with pytest.raises(IndexError):
        matrix[0, -1000] = 0


def test_rows_cols():
    assert zivid.Matrix4x4.rows == 4
    assert zivid.Matrix4x4.cols == 4
    assert zivid.Matrix4x4().rows == 4
    assert zivid.Matrix4x4().cols == 4

    with pytest.raises(AttributeError):
        zivid.Matrix4x4.rows = 3

    with pytest.raises(AttributeError):
        zivid.Matrix4x4.cols = 3


def test_inverse():
    def invertible_matrix():
        return [
            [1, 1, 1, -1],
            [1, 1, -1, 1],
            [1, -1, 1, 1],
            [-1, 1, 1, 1],
        ]

    def non_invertible_matrix():
        return [1] * 16

    matrix = zivid.Matrix4x4(invertible_matrix())

    assert_all_equal_2d(
        matrix.inverse(),
        [
            [0.25, 0.25, 0.25, -0.25],
            [0.25, 0.25, -0.25, 0.25],
            [0.25, -0.25, 0.25, 0.25],
            [-0.25, 0.25, 0.25, 0.25],
        ],
    )

    assert_all_equal_2d(matrix, invertible_matrix())

    matrix = zivid.Matrix4x4(non_invertible_matrix())

    with pytest.raises(RuntimeError):
        matrix.inverse()

    assert_all_equal_flat(matrix, non_invertible_matrix())


def test_buffer_protocol():
    numpy_array = numpy.array(zivid.Matrix4x4(sample_2d_list()))
    assert (numpy_array == sample_2d_list()).all()


def test_to_string():
    matrix = zivid.Matrix4x4(sample_2d_list())

    assert str(matrix) == (
        "[ [-8.300000,  6.750000, -2.000000,  5.700000], \n"
        "  [-0.240000,  1.490000,  3.500000, -4.250000], \n"
        "  [ 52.000000,  0.980000,  970.599976,  75000.000000], \n"
        "  [-64.300003,  15.400000, -84.400002, -13.400000] ]"
    )


def test_save():
    with tempfile.TemporaryDirectory() as tmpdir, zivid.Application() as _:
        file = Path(tmpdir) / "matrix_saved.yml"
        matrix = zivid.Matrix4x4(
            to_float32_2d(
                [
                    [0.5, 1.34, -234, -3.43],
                    [-4.31, 5343, 6.34, 7.12],
                    [-8, -9, -10.3, 11.2],
                    [1.2, 13.5, 1.4, 0.15],
                ]
            )
        )
        assert not file.exists()

        matrix.save(file)

        assert file.exists()

        expected_content = (
            "FloatMatrix:\n"
            "  Data: [\n"
            "    [0.5, 1.34, -234, -3.43],\n"
            "    [-4.31, 5343, 6.34, 7.12],\n"
            "    [-8, -9, -10.3, 11.2],\n"
            "    [1.2, 13.5, 1.4, 0.15]]\n"
        )

        with open(file, "r", encoding="utf8") as file:
            assert expected_content in file.read()


def test_load():
    with tempfile.TemporaryDirectory() as tmpdir, zivid.Application() as _:
        file = Path(tmpdir) / "matrix_saved.yml"
        zivid.Matrix4x4(sample_2d_list()).save(file)
        matrix = zivid.Matrix4x4()
        matrix.load(file)
        assert_all_equal_2d(matrix, sample_2d_list())


def test_file_init():
    with tempfile.TemporaryDirectory() as tmpdir, zivid.Application() as _:
        file = Path(tmpdir) / "matrix_saved.yml"
        zivid.Matrix4x4(sample_2d_list()).save(file)
        assert_all_equal_2d(zivid.Matrix4x4(file), sample_2d_list())


def test_implicit_convert_to_numpy():
    sample = to_float32_2d(
        [
            [1.0, 0.0, 0.0, 10.0],
            [0.0, 0.0, -1.0, 20.0],
            [0.0, 1.0, 0.0, 30.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    pose1 = zivid.calibration.Pose(zivid.Matrix4x4(sample))
    pose2 = zivid.calibration.Pose(numpy.array(sample))

    assert_all_equal_2d(zivid.Matrix4x4(pose1.to_matrix()), pose2.to_matrix())


def test_identity():
    identity = zivid.Matrix4x4.identity()
    assert numpy.array_equal(identity, numpy.identity(4, dtype=numpy.float32))



"""Contains the Capture Assistant functionality."""

import _zivid

from zivid._suggest_settings_parameters import (  # pylint: disable=unused-import
    SuggestSettingsParameters,
)
from zivid._suggest_settings_parameters import (
    _to_internal_capture_assistant_suggest_settings_parameters,
)
from zivid.settings import _to_settings


def suggest_settings(camera, suggest_settings_parameters):
    """Find settings for the current scene based on given parameters.

    The suggested settings returned from this function should be passed into
    camera.capture() to capture and retrieve the Frame containing a point cloud.

    Args:
        camera: A Camera instance
        suggest_settings_parameters: A SuggestSettingsParameters instance

    Returns:
        A Settings instance optimized for the current scene
    """
    internal_settings = _zivid.capture_assistant.suggest_settings(
        camera._Camera__impl,  # pylint: disable=protected-access
        _to_internal_capture_assistant_suggest_settings_parameters(
            suggest_settings_parameters
        ),
    )
    return _to_settings(internal_settings)



import pytest


# These tests cannot be in the "test_application" module because that module depends on a
# zivid.Application fixture and these tests want to test the initialization and teardown of the
# zivid.Application class.


def test_init_with():
    import zivid

    with zivid.Application() as app:
        assert app
        assert isinstance(app, zivid.Application)


def test_release(file_camera_file):
    import zivid

    application = zivid.Application()
    assert application.create_file_camera(file_camera_file)
    application.release()
    with pytest.raises(RuntimeError):
        application.create_file_camera(file_camera_file)



"""Module for calibration features, such as HandEye and MultiCamera."""

# pylint: disable=unused-import
from zivid._calibration.detector import (
    DetectionResult,
    DetectionResultFiducialMarkers,
    MarkerShape,
    MarkerDictionary,
    detect_feature_points,
    detect_calibration_board,
    capture_calibration_board,
    detect_markers,
)
from zivid._calibration.hand_eye import (
    HandEyeInput,
    HandEyeResidual,
    HandEyeOutput,
    calibrate_eye_in_hand,
    calibrate_eye_to_hand,
)
from zivid._calibration.multi_camera import (
    MultiCameraResidual,
    MultiCameraOutput,
    calibrate_multi_camera,
)
from zivid._calibration.pose import Pose



"""Module containing implementation of feature point detection functionality.

This module should not be imported directly by end-user, but rather accessed through
the zivid.calibration module.
"""

import numpy

import _zivid
from zivid.camera import Camera
from zivid.frame import Frame
from zivid._calibration.pose import Pose


class DetectionResult:
    """Class representing detected feature points from a calibration board."""

    def __init__(self, impl):
        """Initialize DetectionResult wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if not isinstance(impl, _zivid.calibration.DetectionResult):
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}".format(
                    type(impl), _zivid.calibration.DetectionResult
                )
            )

        self.__impl = impl

    def valid(self):
        """Check validity of DetectionResult.

        Returns:
            True if DetectionResult is valid
        """
        return self.__impl.valid()

    def centroid(self):
        """Get the centroid of the detected feature points.

        Will throw an exception if the DetectionResult is not valid.

        Returns:
            A 1D array containing the X, Y and Z coordinates of the centroid
        """
        return self.__impl.centroid()

    def pose(self):
        """Get position and orientation of the top left detected corner in camera-space.

        Pose calculation works for official Zivid calibration boards only.
        An exception will be thrown if valid() is false or if the board is not supported.

        Returns:
            The Pose of the top left corner (4x4 transformation matrix)
        """
        return Pose(self.__impl.pose().to_matrix())

    def __bool__(self):
        return bool(self.__impl)

    def __str__(self):
        return str(self.__impl)


class MarkerShape:
    """Holds physical (3D) and image (2D) properties of a detected fiducial marker."""

    def __init__(self, impl):
        """Initialize MarkerShape wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if not isinstance(impl, _zivid.calibration.MarkerShape):
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}".format(
                    type(impl), _zivid.calibration.MarkerShape
                )
            )

        self.__impl = impl

    @property
    def corners_in_pixel_coordinates(self):
        """Get 2D image coordinates of the corners of the detected marker.

        Returns:
            Four 2D corner coordinates as a 4x2 numpy array
        """
        return numpy.array(self.__impl.corners_in_pixel_coordinates())

    @property
    def corners_in_camera_coordinates(self):
        """Get 3D spatial coordinates of the corners of the detected marker.

        Returns:
            Four 3D corner coordinates as a 4x3 numpy array
        """
        return numpy.array(self.__impl.corners_in_camera_coordinates())

    @property
    def identifier(self):
        """Get the id of the detected marker.

        Returns:
            Id as int
        """
        return self.__impl.id_()

    @property
    def pose(self):
        """Get 3D pose of the marker.

        The returned pose will be positioned at the center of the marker, and have an orientation such that its z-axis
        points perpendicularly into the face of the marker.

        Returns:
            The Pose of the marker center (4x4 transformation matrix)
        """
        return Pose(self.__impl.pose().to_matrix())


class MarkerDictionary:
    """Holds information about fiducial markers such as ArUco or AprilTag for detection.

    This class's properties describe the different dictionaries available, for example
    aruco4x4_50 describes the ArUco dictionary with 50 markers of size 4x4.

    For more information on ArUco markers see the OpenCV documentation on ArUco markers:
    https://docs.opencv.org/4.x/d5/dae/tutorial_aruco_detection.html,

    To get more information about fiducial markers in general, refer to the wikipedia page:
    https://en.wikipedia.org/wiki/Fiducial_marker
    """

    aruco4x4_50 = "aruco4x4_50"
    aruco4x4_100 = "aruco4x4_100"
    aruco4x4_250 = "aruco4x4_250"
    aruco4x4_1000 = "aruco4x4_1000"
    aruco5x5_50 = "aruco5x5_50"
    aruco5x5_100 = "aruco5x5_100"
    aruco5x5_250 = "aruco5x5_250"
    aruco5x5_1000 = "aruco5x5_1000"
    aruco6x6_50 = "aruco6x6_50"
    aruco6x6_100 = "aruco6x6_100"
    aruco6x6_250 = "aruco6x6_250"
    aruco6x6_1000 = "aruco6x6_1000"
    aruco7x7_50 = "aruco7x7_50"
    aruco7x7_100 = "aruco7x7_100"
    aruco7x7_250 = "aruco7x7_250"
    aruco7x7_1000 = "aruco7x7_1000"

    _valid_values = {
        "aruco4x4_50": _zivid.calibration.MarkerDictionary.aruco4x4_50,
        "aruco4x4_100": _zivid.calibration.MarkerDictionary.aruco4x4_100,
        "aruco4x4_250": _zivid.calibration.MarkerDictionary.aruco4x4_250,
        "aruco4x4_1000": _zivid.calibration.MarkerDictionary.aruco4x4_1000,
        "aruco5x5_50": _zivid.calibration.MarkerDictionary.aruco5x5_50,
        "aruco5x5_100": _zivid.calibration.MarkerDictionary.aruco5x5_100,
        "aruco5x5_250": _zivid.calibration.MarkerDictionary.aruco5x5_250,
        "aruco5x5_1000": _zivid.calibration.MarkerDictionary.aruco5x5_1000,
        "aruco6x6_50": _zivid.calibration.MarkerDictionary.aruco6x6_50,
        "aruco6x6_100": _zivid.calibration.MarkerDictionary.aruco6x6_100,
        "aruco6x6_250": _zivid.calibration.MarkerDictionary.aruco6x6_250,
        "aruco6x6_1000": _zivid.calibration.MarkerDictionary.aruco6x6_1000,
        "aruco7x7_50": _zivid.calibration.MarkerDictionary.aruco7x7_50,
        "aruco7x7_100": _zivid.calibration.MarkerDictionary.aruco7x7_100,
        "aruco7x7_250": _zivid.calibration.MarkerDictionary.aruco7x7_250,
        "aruco7x7_1000": _zivid.calibration.MarkerDictionary.aruco7x7_1000,
    }

    @classmethod
    def valid_values(cls):
        """Get valid values for MarkerDictionary.

        Returns:
            A list of strings representing valid values for MarkerDictionary.
        """
        return list(cls._valid_values.keys())

    @classmethod
    def marker_count(cls, dictionary_name):
        """Get the number of markers in a dictionary.

        Args:
            dictionary_name: Name of the dictionary, e.g. "aruco4x4_50". Must be one of the values returned by
                valid_values().

        Returns:
            Number of markers in the dictionary.

        Raises:
            ValueError: If the dictionary name is not one of the valid values returned by
                valid_values().
        """
        if dictionary_name not in cls._valid_values:
            raise ValueError(
                "Invalid dictionary name '{}'. Valid values are {}".format(
                    dictionary_name, cls.valid_values()
                )
            )

        return cls._valid_values[dictionary_name].marker_count()


class DetectionResultFiducialMarkers:
    """Class representing detected fiducial markers."""

    def __init__(self, impl):
        """Initialize DetectionResultFiducialMarkers wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if not isinstance(impl, _zivid.calibration.DetectionResultFiducialMarkers):
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}".format(
                    type(impl), _zivid.calibration.DetectionResultFiducialMarkers
                )
            )

        self.__impl = impl

    def valid(self):
        """Check validity of DetectionResult.

        Returns:
            True if DetectionResult is valid
        """
        return self.__impl.valid()

    def allowed_marker_ids(self):
        """Get the allowed marker ids this detection result was made with.

        Returns:
            A list of integers, equal to what was passed to the detection function.
        """
        return self.__impl.allowed_marker_ids()

    def detected_markers(self):
        """Get all detected markers.

        Returns:
            A list of MarkerShape instances
        """
        return [MarkerShape(impl) for impl in self.__impl.detected_markers()]

    def __bool__(self):
        return bool(self.__impl)

    def __str__(self):
        return str(self.__impl)


def detect_feature_points(point_cloud):
    """Detect feature points from a calibration object in a point cloud.

    Args:
        point_cloud: PointCloud containing a calibration object

    Returns:
        A DetectionResult instance
    """

    return DetectionResult(
        _zivid.calibration.detect_feature_points(
            point_cloud._PointCloud__impl  # pylint: disable=protected-access
        )
    )


def detect_calibration_board(source):
    """
    Detect feature points from a calibration board in a frame or using a given camera.

    If a camera is used, this function will perform a relatively slow but high-quality point cloud
    capture with the camera. This function is necessary for applications that require very
    high-accuracy DetectionResult, such as in-field verification/correction.

    The functionality is to be exclusively used in combination with Zivid verified calibration boards.
    For further information please visit https://support.zivid.com.

    Args:
        source: A frame containing an image of a calibration board or a camera pointed at
            a calibration board

    Raises:
        TypeError: If source is not of type Camera or Frame

    Returns:
        A DetectionResult instance
    """

    if isinstance(source, Camera):
        return DetectionResult(
            _zivid.calibration.detect_calibration_board(
                source._Camera__impl  # pylint: disable=protected-access
            )
        )
    if isinstance(source, Frame):
        return DetectionResult(
            _zivid.calibration.detect_calibration_board(
                source._Frame__impl  # pylint: disable=protected-access
            )
        )
    raise TypeError(
        "Unsupported type for argument source. Got {}, expected one of {}".format(
            type(source),
            (Camera, Frame),
        )
    )


def capture_calibration_board(camera):
    """
    Capture a calibration board with the given camera.

    The functionality is to be exclusively used in combination with Zivid verified calibration boards.
    For further information please visit https://support.zivid.com.

    This function will perform a relatively slow but high-quality point cloud capture with the
    given camera. This function is necessary for applications that require very high-accuracy
    captures, such as in-field verification/correction.

    The Frame that is returned from this function may be used as input to `detect_calibration_board`.
    You may also use `detect_calibration_board` directly, which will invoke this function under the
    hood and yield a DetectionResult.

    Args:
        camera: a Camera pointed at a calibration board

    Returns:
        A Frame
    """
    return Frame(
        _zivid.calibration.capture_calibration_board(
            camera._Camera__impl  # pylint: disable=protected-access
        )
    )


def detect_markers(frame, allowed_marker_ids, marker_dictionary):
    """Detect fiducial markers such as ArUco markers in a frame.

    Only markers with integer IDs are supported. To get more information about fiducial markers, refer to the
    wikipedia page:  https://en.wikipedia.org/wiki/Fiducial_marker

    For more information on ArUco markers specifically, see the OpenCV documentation on ArUco markers:
    https://docs.opencv.org/4.x/d5/dae/tutorial_aruco_detection.html,

    Frame need not contain all markers listed in allowedMarkerIds for a successful detection.

    Args:
        frame: A frame containing an image of one or several fiducial markers
        allowed_marker_ids: List of the IDs of markers to be detected
        marker_dictionary: The name of the marker dictionary to use. The name must be one of the values returned by
            MarkerDictionary.valid_values()

    Raises:
        ValueError: If marker_dictionary is not one of the valid values returned by MarkerDictionary.valid_values()

    Returns:
        A DetectionResultFiducialMarkers instance
    """

    if marker_dictionary not in MarkerDictionary.valid_values():
        raise ValueError(
            "Invalid marker dictionary '{}'. Valid values are {}".format(
                marker_dictionary, MarkerDictionary.valid_values()
            )
        )
    dictionary = MarkerDictionary._valid_values.get(  # pylint: disable=protected-access
        marker_dictionary
    )

    return DetectionResultFiducialMarkers(
        _zivid.calibration.detect_markers(
            frame._Frame__impl,  # pylint: disable=protected-access
            allowed_marker_ids,
            dictionary,
        )
    )



"""Contains the Preset and Category classes."""

import _zivid

from zivid.settings import _to_settings
from zivid.camera_info import CameraInfo, _to_internal_camera_info
from zivid.settings2d import _to_settings2d


class Preset:
    """Settings preset.

    Presets are pre-defined settings that are tuned for different camera models to perform optimally
    under different conditions and use cases.

    This class cannot be initialized directly by the end-user. Use the presets method on the
    Category class to obtain a Preset instance.
    """

    def __init__(self, impl):
        """Initialize Preset wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if not any(
            [
                isinstance(impl, _zivid.presets.Preset),
                isinstance(impl, _zivid.presets.Preset2D),
            ]
        ):
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}".format(
                    type(impl),
                    " or ".join(
                        (
                            _zivid.presets.Preset.__name__,
                            _zivid.presets.Preset2D.__name__,
                        )
                    ),
                )
            )
        self.__impl = impl

    @property
    def name(self):
        """Get the name of the preset.

        Returns:
            The name of the preset.
        """
        return self.__impl.name()

    @property
    def settings(self):
        """Get the settings of the preset.

        The settings might change between different releases of the SDK. New presets might be added
        and old ones might be removed. If having the exact same settings is desired, it is
        recommended to save them to a YML file and load as needed.

        Returns:
                The settings of the preset. The settings will be an instance of Settings if the preset is a 3D preset,
                or an instance of Settings2D if the preset is a 2D preset.
        """

        settings = self.__impl.settings()

        if isinstance(settings, _zivid.Settings2D):
            return _to_settings2d(settings)

        return _to_settings(settings)

    def __str__(self):
        return str(self.__impl)


class Category:
    """Preset category.

    This class cannot be initialized directly by the end-user. Use the categories or categories2d functions to obtain a
    list of categories available for a camera.
    """

    def __init__(self, impl):
        """Initialize Category wrapper.

        This constructor is only used internally, and should not be called by the end-user.

        Args:
            impl:   Reference to internal/back-end instance.

        Raises:
            TypeError: If argument does not match the expected internal class.
        """
        if not any(
            [
                isinstance(impl, _zivid.presets.Category),
                isinstance(impl, _zivid.presets.Category2D),
            ]
        ):
            raise TypeError(
                "Unsupported type for argument impl. Got {}, expected {}".format(
                    type(impl),
                    " or ".join(
                        (
                            _zivid.presets.Category.__name__,
                            _zivid.presets.Category2D.__name__,
                        )
                    ),
                )
            )
        self.__impl = impl

    @property
    def name(self):
        """Get the name of the category.

        Returns:
            The name of the category.
        """
        return self.__impl.name()

    @property
    def presets(self):
        """Get the presets available in the category.

        The settings might change between different releases of the SDK. New presets might be added
        and old ones might be removed. If having the exact same settings is desired, it is
        recommended to save them to a YML file and load as needed.

        Returns:
            The presets available in the category.
        """
        return [Preset(p) for p in self.__impl.presets()]

    def __str__(self):
        return str(self.__impl)


def categories(model):
    """Get available preset categories for the specified camera model.

    A preset category contains a collection of presets optimized for one scenario or use case.

    The settings might change between different releases of the SDK. New presets might be added
    and old ones might be removed. If having the exact same settings is desired, it is recommended
    to save them to a YML file and load as needed.

    Args:
        model: The model for the camera whose preset categories should be returned. This value can
               be obtained from CameraInfo.model.

    Returns:
        The available categories for the specified camera model.
    """
    return [
        Category(c)
        for c in _zivid.presets.categories(
            _to_internal_camera_info(CameraInfo(model=model)).model
        )
    ]


def categories2d(model):
    """Get available 2D preset categories for the specified camera model.

    See `categories` for more information.

    Args:
        model: The model for the camera whose 2D preset categories should be returned. This value can
               be obtained from CameraInfo.model.

    Returns:
        The available 2D categories for the specified camera model.
    """
    return [
        Category(c)
        for c in _zivid.presets.categories2d(
            _to_internal_camera_info(CameraInfo(model=model)).model
        )
    ]



import pytest


def test_point_cloud_copy_data(point_cloud):
    import numpy as np

    # Copy all possible formats
    xyz = point_cloud.copy_data("xyz")
    xyzw = point_cloud.copy_data("xyzw")
    xyzrgba = point_cloud.copy_data("xyzrgba")
    xyzbgra = point_cloud.copy_data("xyzbgra")
    rgba = point_cloud.copy_data("rgba")
    bgra = point_cloud.copy_data("bgra")
    srgb = point_cloud.copy_data("srgb")
    normals = point_cloud.copy_data("normals")
    depth = point_cloud.copy_data("z")
    snr = point_cloud.copy_data("snr")
    assert isinstance(xyz, np.ndarray)
    assert isinstance(xyzw, np.ndarray)
    assert isinstance(xyzrgba, np.ndarray)
    assert isinstance(xyzbgra, np.ndarray)
    assert isinstance(rgba, np.ndarray)
    assert isinstance(bgra, np.ndarray)
    assert isinstance(srgb, np.ndarray)
    assert isinstance(normals, np.ndarray)
    assert isinstance(depth, np.ndarray)
    assert isinstance(snr, np.ndarray)

    # Check errors when argument is wrong or missing
    with pytest.raises(ValueError):
        point_cloud.copy_data("bogus-format")
    with pytest.raises(TypeError):
        point_cloud.copy_data()


def test_point_cloud_xyzw(point_cloud):
    import numpy as np

    xyz = point_cloud.copy_data("xyz")
    xyzw = point_cloud.copy_data("xyzw")
    xyzrgba = point_cloud.copy_data("xyzrgba")
    xyzbgra = point_cloud.copy_data("xyzbgra")
    depth = point_cloud.copy_data("z")

    assert depth.shape == (point_cloud.height, point_cloud.width)
    assert xyz.shape == (point_cloud.height, point_cloud.width, 3)
    assert xyzw.shape == (point_cloud.height, point_cloud.width, 4)
    assert depth.dtype == np.float32
    assert xyz.dtype == np.float32
    assert xyzw.dtype == np.float32
    assert xyzrgba["x"].dtype == np.float32
    assert xyzrgba["y"].dtype == np.float32
    assert xyzrgba["z"].dtype == np.float32
    assert xyzbgra["x"].dtype == np.float32
    assert xyzbgra["y"].dtype == np.float32
    assert xyzbgra["z"].dtype == np.float32
    np.testing.assert_array_equal(xyz[:, :, 2], depth)
    np.testing.assert_array_equal(xyz[:, :, 0], xyzw[:, :, 0])
    np.testing.assert_array_equal(xyz[:, :, 1], xyzw[:, :, 1])
    np.testing.assert_array_equal(xyz[:, :, 2], xyzw[:, :, 2])


def test_point_cloud_rgba(point_cloud):
    import numpy as np

    xyzrgba = point_cloud.copy_data("xyzrgba")
    xyzbgra = point_cloud.copy_data("xyzbgra")
    rgba = point_cloud.copy_data("rgba")
    bgra = point_cloud.copy_data("bgra")
    srgb = point_cloud.copy_data("srgb")

    assert rgba.shape == (point_cloud.height, point_cloud.width, 4)
    assert bgra.shape == (point_cloud.height, point_cloud.width, 4)
    assert srgb.shape == (point_cloud.height, point_cloud.width, 4)
    assert xyzrgba.shape == (point_cloud.height, point_cloud.width)
    assert xyzbgra.shape == (point_cloud.height, point_cloud.width)
    assert rgba.dtype == np.uint8
    assert bgra.dtype == np.uint8
    assert srgb.dtype == np.uint8
    assert xyzrgba["r"].dtype == np.uint8
    assert xyzrgba["g"].dtype == np.uint8
    assert xyzrgba["b"].dtype == np.uint8
    assert xyzrgba["a"].dtype == np.uint8
    np.testing.assert_array_equal(rgba[:, :, 0], xyzrgba["r"])
    np.testing.assert_array_equal(rgba[:, :, 1], xyzrgba["g"])
    np.testing.assert_array_equal(rgba[:, :, 2], xyzrgba["b"])
    np.testing.assert_array_equal(rgba[:, :, 3], xyzrgba["a"])
    assert xyzbgra["b"].dtype == np.uint8
    assert xyzbgra["g"].dtype == np.uint8
    assert xyzbgra["r"].dtype == np.uint8
    assert xyzbgra["a"].dtype == np.uint8
    np.testing.assert_array_equal(bgra[:, :, 0], xyzbgra["b"])
    np.testing.assert_array_equal(bgra[:, :, 1], xyzbgra["g"])
    np.testing.assert_array_equal(bgra[:, :, 2], xyzbgra["r"])
    np.testing.assert_array_equal(bgra[:, :, 3], xyzbgra["a"])
    np.testing.assert_array_equal(xyzbgra["r"], xyzrgba["r"])
    np.testing.assert_array_equal(xyzbgra["g"], xyzrgba["g"])
    np.testing.assert_array_equal(xyzbgra["b"], xyzrgba["b"])
    np.testing.assert_array_equal(xyzbgra["a"], xyzrgba["a"])
    np.testing.assert_array_equal(bgra[:, :, 0], rgba[:, :, 2])
    np.testing.assert_array_equal(bgra[:, :, 1], rgba[:, :, 1])
    np.testing.assert_array_equal(bgra[:, :, 2], rgba[:, :, 0])
    np.testing.assert_array_equal(bgra[:, :, 3], rgba[:, :, 3])


def test_point_cloud_copy_image(point_cloud):
    import numpy as np
    import zivid

    image_rgba = point_cloud.copy_image("rgba")
    assert isinstance(image_rgba, zivid.Image)
    assert image_rgba.height == point_cloud.height
    assert image_rgba.width == point_cloud.width

    image_bgra = point_cloud.copy_image("bgra")
    assert isinstance(image_bgra, zivid.Image)
    assert image_bgra.height == point_cloud.height
    assert image_bgra.width == point_cloud.width

    image_srgb = point_cloud.copy_image("srgb")
    assert isinstance(image_srgb, zivid.Image)
    assert image_srgb.height == point_cloud.height
    assert image_srgb.width == point_cloud.width

    rgba = point_cloud.copy_data("rgba")
    np.testing.assert_array_equal(image_rgba.copy_data(), rgba)

    bgra = point_cloud.copy_data("bgra")
    np.testing.assert_array_equal(image_bgra.copy_data(), bgra)

    srgb = point_cloud.copy_data("srgb")
    np.testing.assert_array_equal(image_srgb.copy_data(), srgb)

    # Check errors when argument is wrong or missing
    with pytest.raises(ValueError):
        point_cloud.copy_image("bogus-format")
    with pytest.raises(TypeError):
        point_cloud.copy_image()


def test_point_cloud_normals(point_cloud):
    import numpy as np

    normals = point_cloud.copy_data("normals")

    assert normals.shape == (point_cloud.height, point_cloud.width, 3)
    assert normals.dtype == np.float32

    normals_flat = normals.reshape(normals.shape[0] * normals.shape[1], 3)
    non_nan_normals = normals_flat[~np.isnan(normals_flat[:, 0])]
    vector_lengths = np.linalg.norm(non_nan_normals, axis=1)
    np.testing.assert_array_almost_equal(vector_lengths, 1.0)


def test_point_cloud_snr(point_cloud):
    import numpy as np

    snr = point_cloud.copy_data("snr")

    assert snr.dtype == np.float32
    assert snr.shape == (point_cloud.height, point_cloud.width)
    assert np.all(snr >= 0.0)
    assert np.all(snr < 1000)


def test_height(point_cloud):
    height = point_cloud.height

    assert height is not None
    assert isinstance(height, int)


def test_width(point_cloud):
    width = point_cloud.width

    assert width is not None
    assert isinstance(width, int)


def _validate_transformation(xyzw_before, xyzw_after, transform):
    import numpy as np

    # Pick an arbitary point to test
    i = xyzw_before.shape[0] // 3
    j = xyzw_before.shape[1] // 3
    point_before = xyzw_before[i, j, :]
    point_after = xyzw_after[i, j, :]
    assert np.all(~np.isnan(point_after))
    assert np.all(~np.isnan(point_before))
    point_after_expected = np.dot(transform, point_before)
    np.testing.assert_allclose(point_after, point_after_expected, rtol=1e-6)


def test_transform(point_cloud, transform):
    import zivid

    # Get points before and after transform
    xyzw_before = point_cloud.copy_data("xyzw")
    point_cloud_returned = point_cloud.transform(transform)
    xyzw_after = point_cloud.copy_data("xyzw")

    # Check that return value is just a reference to the original object
    assert isinstance(point_cloud_returned, zivid.PointCloud)
    assert point_cloud_returned is point_cloud

    # Check that the transformation was actually applied
    _validate_transformation(xyzw_before, xyzw_after, transform)


def test_transform_chaining(point_cloud, transform):
    import zivid
    import numpy as np

    # Get points before and after transform
    xyzw_before = point_cloud.copy_data("xyzw")
    point_cloud_returned = point_cloud.transform(transform).transform(transform)
    xyzw_after = point_cloud.copy_data("xyzw")

    # Check that return value is just a reference to the original object
    assert isinstance(point_cloud_returned, zivid.PointCloud)
    assert point_cloud_returned is point_cloud

    # Check that the transformation was actually applied
    _validate_transformation(xyzw_before, xyzw_after, np.dot(transform, transform))


def test_downsampling_enum():
    import zivid

    vals = zivid.PointCloud.Downsampling.valid_values()
    assert len(vals) == 3
    assert "by2x2" in vals
    assert "by3x3" in vals
    assert "by4x4" in vals
    assert zivid.PointCloud.Downsampling.by2x2 == "by2x2"
    assert zivid.PointCloud.Downsampling.by3x3 == "by3x3"
    assert zivid.PointCloud.Downsampling.by4x4 == "by4x4"


def _make_downsampling_enum(fraction):
    return "by{f}x{f}".format(f=fraction)


@pytest.mark.parametrize("fraction", [2, 3, 4])
def test_downsample(point_cloud, fraction):
    import zivid

    # Remember original size
    height_orig = point_cloud.height
    width_orig = point_cloud.width

    # Perform downsampling
    point_cloud_returned = point_cloud.downsample(_make_downsampling_enum(fraction))

    # Check that return value is just a reference to the original object
    assert isinstance(point_cloud_returned, zivid.PointCloud)
    assert point_cloud_returned is point_cloud

    # Check that the new size is as expected
    assert height_orig // fraction == point_cloud.height
    assert width_orig // fraction == point_cloud.width


@pytest.mark.parametrize("fraction", [2, 3, 4])
def test_downsampled(point_cloud, fraction):
    import zivid

    # Remember original size
    height_orig = point_cloud.height
    width_orig = point_cloud.width

    # Perform downsampling
    point_cloud_new = point_cloud.downsampled(_make_downsampling_enum(fraction))

    # Check that a new object was returned and that the original is untouched
    assert isinstance(point_cloud_new, zivid.PointCloud)
    assert point_cloud_new is not point_cloud
    assert point_cloud.height == height_orig
    assert point_cloud.width == width_orig

    # Check that the new size is as expected
    assert height_orig // fraction == point_cloud_new.height
    assert width_orig // fraction == point_cloud_new.width


def test_downsample_chaining(point_cloud):
    import zivid

    # Remember original size
    height_orig = point_cloud.height
    width_orig = point_cloud.width

    # Perform downsampling
    point_cloud_returned = point_cloud.downsample("by2x2").downsample("by3x3")

    # Check that return value is just a reference to the original object
    assert isinstance(point_cloud_returned, zivid.PointCloud)
    assert point_cloud_returned is point_cloud

    # Check that the new size is as expected
    assert height_orig // 2 // 3 == point_cloud.height
    assert width_orig // 2 // 3 == point_cloud.width


def test_downsampled_chaining(point_cloud):
    import zivid

    # Remember original size
    height_orig = point_cloud.height
    width_orig = point_cloud.width

    # Perform downsampling
    point_cloud_new = point_cloud.downsampled("by2x2").downsampled("by3x3")

    # Check that a new object was returned and that the original is untouched
    assert isinstance(point_cloud_new, zivid.PointCloud)
    assert point_cloud_new is not point_cloud
    assert point_cloud.height == height_orig
    assert point_cloud.width == width_orig

    # Check result
    assert height_orig // 2 // 3 == point_cloud_new.height
    assert width_orig // 2 // 3 == point_cloud_new.width


def test_height_context_manager(frame):
    with frame.point_cloud() as point_cloud:
        point_cloud.height
    with pytest.raises(RuntimeError):
        point_cloud.height


def test_width_context_manager(frame):
    with frame.point_cloud() as point_cloud:
        point_cloud.width
    with pytest.raises(RuntimeError):
        point_cloud.width


def test_copy_data_context_manager(frame):
    with frame.point_cloud() as point_cloud:
        point_cloud.copy_data(data_format="xyzrgba")
    with pytest.raises(RuntimeError):
        point_cloud.copy_data(data_format="xyzrgba")
    with pytest.raises(RuntimeError):
        point_cloud.copy_data(data_format=123)
    with pytest.raises(TypeError):
        point_cloud.copy_data()


def test_illegal_init(application):
    import zivid

    with pytest.raises(TypeError):
        zivid.PointCloud()  # pylint: disable=no-value-for-parameter

    with pytest.raises(TypeError):
        zivid.PointCloud("Should fail.")

    with pytest.raises(TypeError):
        zivid.PointCloud(123)



import tempfile
import platform
import subprocess
from sys import version_info
from tempfile import TemporaryDirectory
from pathlib import Path
from pkgutil import iter_modules
from skbuild import setup, constants


# To be replaced by: from setuptools_scm import get_version
def get_version():
    return "2.14.2"


def _zivid_sdk_version():
    return "2.14.2"


def _zivid_python_version():
    scm_version = get_version()

    if "+" in scm_version:
        base_version, scm_metadata = scm_version.split("+", 1)
    else:
        base_version = scm_version
        scm_metadata = None

    base_version = "{}.{}".format(base_version, _zivid_sdk_version())

    if scm_metadata:
        version = "{}+{}".format(base_version, scm_metadata)
    else:
        version = base_version

    return version


def _python_version():
    return "{}.{}.{}".format(*version_info)


def _make_message_box(*message):
    width = max(len(e) for e in message)

    box_bar = "+-" + "-" * width + "-+"
    empty_line = "\n| " + " " * width + " |\n"
    message_lines = ["| " + line + " " * (width - len(line)) + " |" for line in message]

    return (
        "\n\n" + box_bar + "\n" + empty_line.join(message_lines) + "\n" + box_bar + "\n"
    )


def _check_dependency(module_name, package_hint=None):
    if package_hint is None:
        package_hint = module_name
    if module_name not in [module[1] for module in iter_modules()]:
        raise ImportError(
            _make_message_box(
                "!! Missing module '{}' !!".format(module_name),
                "Please install '{}' manually or use PIP>=19 to handle build dependencies automatically (PEP 517)".format(
                    package_hint
                ),
            )
        )


def _check_cpp17_compiler():
    def run_process(args, **kwargs):
        try:
            with subprocess.Popen(args, **kwargs) as process:
                exit_code = process.wait()
                if exit_code != 0:
                    raise RuntimeError(
                        "Wait failed with exit code {}".format(exit_code)
                    )
        except Exception as ex:
            raise type(ex)("Process failed: '{}'.".format(" ".join(args))) from ex

    try:
        run_process(("cmake", "--version"))
    except Exception as ex:
        raise RuntimeError(
            _make_message_box(
                "!! CMake not found !!",
                "Ensure PIP is updated to the latest version to handle build dependencies automatically,",
                "or install CMake and Ninja manually.",
            )
        ) from ex
    with tempfile.TemporaryDirectory(prefix="zivid-python-build-") as temp_dir:
        with (Path(temp_dir) / "lib.cpp").open("w") as lib_cpp:
            # MSVC does not report itself as C++17, on Windoes we have to rely on the CMAKE_CXX_STANDARD test below
            if platform.system() == "Linux":
                lib_cpp.write("static_assert(__cplusplus >= 201703L);")
        with (Path(temp_dir) / "CMakeLists.txt").open("w") as cmake_lists_txt:
            cmake_lists_txt.write(
                "project(zivid-python-compiler-detection LANGUAGES CXX)\n"
                "set(CMAKE_CXX_STANDARD 17)\n"
                "add_library(lib lib.cpp)\n"
            )
        try:
            if platform.system() == "Linux":
                run_process(("cmake", "-GNinja", "."), cwd=temp_dir)
            else:
                run_process(("cmake", "."), cwd=temp_dir)
            run_process(("cmake", "--build", "."), cwd=temp_dir)
        except Exception as ex:
            raise RuntimeError(
                _make_message_box(
                    "!! Module setup failed !!",
                    "Make sure you have a working C++17 compiler installed",
                    "Refer to Readme.md for detailed installation instructions",
                )
            ) from ex


def _main():
    # This list is a duplicate of the build-system requirements in pyproject.toml.
    # The purpose of these checks is to help users with PIP<19 lacking support for
    # pyproject.toml
    # Keep the two lists in sync
    _check_dependency("cmake")
    _check_dependency("conan")
    _check_dependency("ninja")
    _check_dependency("skbuild", "scikit-build")

    _check_cpp17_compiler()

    with TemporaryDirectory(prefix="zivid-python-build_") as build_dir:
        print("Overriding build dir: " + build_dir)
        constants.SKBUILD_DIR = lambda: build_dir

        setup(
            name="zivid",
            version=_zivid_python_version(),
            description="Defining the Future of 3D Machine Vision",
            long_description=Path("README.md").read_text(encoding="utf-8"),
            long_description_content_type="text/markdown",
            url="https://www.zivid.com",
            author="Zivid AS",
            author_email="customersuccess@zivid.com",
            license="BSD 3-Clause",
            packages=[
                "zivid",
                "zivid._calibration",
                "zivid.experimental",
                "zivid.experimental.point_cloud_export",
                "_zivid",
            ],
            package_dir={"": "modules"},
            install_requires=["numpy"],
            cmake_args=[
                "-DZIVID_PYTHON_VERSION=" + _zivid_python_version(),
                "-DZIVID_SDK_VERSION=" + _zivid_sdk_version(),
                "-DPYTHON_INTERPRETER_VERSION=" + _python_version(),
            ],
            classifiers=[
                "License :: OSI Approved :: BSD License",
                "Programming Language :: Python :: 3",
                "Programming Language :: Python :: Implementation :: CPython",
                "Topic :: Scientific/Engineering",
            ],
        )


if __name__ == "__main__":
    _main()



"""Auto generated, do not edit."""

# pylint: disable=too-many-lines,protected-access,too-few-public-methods,too-many-arguments,line-too-long,missing-function-docstring,missing-class-docstring,redefined-builtin,too-many-branches,too-many-boolean-expressions
import _zivid


class SceneConditions:

    class AmbientLight:

        class FlickerClassification:

            grid50hz = "grid50hz"
            grid60hz = "grid60hz"
            noFlicker = "noFlicker"
            unknownFlicker = "unknownFlicker"

            _valid_values = {
                "grid50hz": _zivid.SceneConditions.AmbientLight.FlickerClassification.grid50hz,
                "grid60hz": _zivid.SceneConditions.AmbientLight.FlickerClassification.grid60hz,
                "noFlicker": _zivid.SceneConditions.AmbientLight.FlickerClassification.noFlicker,
                "unknownFlicker": _zivid.SceneConditions.AmbientLight.FlickerClassification.unknownFlicker,
            }

            @classmethod
            def valid_values(cls):
                return list(cls._valid_values.keys())

        def __init__(
            self,
            flicker_classification=_zivid.SceneConditions.AmbientLight.FlickerClassification().value,
        ):

            if isinstance(
                flicker_classification,
                _zivid.SceneConditions.AmbientLight.FlickerClassification.enum,
            ):
                self._flicker_classification = (
                    _zivid.SceneConditions.AmbientLight.FlickerClassification(
                        flicker_classification
                    )
                )
            elif isinstance(flicker_classification, str):
                self._flicker_classification = (
                    _zivid.SceneConditions.AmbientLight.FlickerClassification(
                        self.FlickerClassification._valid_values[flicker_classification]
                    )
                )
            else:
                raise TypeError(
                    "Unsupported type, expected: str, got {value_type}".format(
                        value_type=type(flicker_classification)
                    )
                )

        @property
        def flicker_classification(self):
            if self._flicker_classification.value is None:
                return None
            for key, internal_value in self.FlickerClassification._valid_values.items():
                if internal_value == self._flicker_classification.value:
                    return key
            raise ValueError(
                "Unsupported value {value}".format(value=self._flicker_classification)
            )

        @flicker_classification.setter
        def flicker_classification(self, value):
            if isinstance(value, str):
                self._flicker_classification = (
                    _zivid.SceneConditions.AmbientLight.FlickerClassification(
                        self.FlickerClassification._valid_values[value]
                    )
                )
            elif isinstance(
                value, _zivid.SceneConditions.AmbientLight.FlickerClassification.enum
            ):
                self._flicker_classification = (
                    _zivid.SceneConditions.AmbientLight.FlickerClassification(value)
                )
            else:
                raise TypeError(
                    "Unsupported type, expected: str, got {value_type}".format(
                        value_type=type(value)
                    )
                )

        def __eq__(self, other):
            if self._flicker_classification == other._flicker_classification:
                return True
            return False

        def __str__(self):
            return str(_to_internal_scene_conditions_ambient_light(self))

    def __init__(
        self,
        ambient_light=None,
    ):

        if ambient_light is None:
            ambient_light = self.AmbientLight()
        if not isinstance(ambient_light, self.AmbientLight):
            raise TypeError(
                "Unsupported type: {value}".format(value=type(ambient_light))
            )
        self._ambient_light = ambient_light

    @property
    def ambient_light(self):
        return self._ambient_light

    @ambient_light.setter
    def ambient_light(self, value):
        if not isinstance(value, self.AmbientLight):
            raise TypeError("Unsupported type {value}".format(value=type(value)))
        self._ambient_light = value

    @classmethod
    def load(cls, file_name):
        return _to_scene_conditions(_zivid.SceneConditions(str(file_name)))

    def save(self, file_name):
        _to_internal_scene_conditions(self).save(str(file_name))

    @classmethod
    def from_serialized(cls, value):
        return _to_scene_conditions(_zivid.SceneConditions.from_serialized(str(value)))

    def serialize(self):
        return _to_internal_scene_conditions(self).serialize()

    def __eq__(self, other):
        if self._ambient_light == other._ambient_light:
            return True
        return False

    def __str__(self):
        return str(_to_internal_scene_conditions(self))


def _to_scene_conditions_ambient_light(internal_ambient_light):
    return SceneConditions.AmbientLight(
        flicker_classification=internal_ambient_light.flicker_classification.value,
    )


def _to_scene_conditions(internal_scene_conditions):
    return SceneConditions(
        ambient_light=_to_scene_conditions_ambient_light(
            internal_scene_conditions.ambient_light
        ),
    )


def _to_internal_scene_conditions_ambient_light(ambient_light):
    internal_ambient_light = _zivid.SceneConditions.AmbientLight()

    internal_ambient_light.flicker_classification = (
        _zivid.SceneConditions.AmbientLight.FlickerClassification(
            ambient_light._flicker_classification.value
        )
    )

    return internal_ambient_light


def _to_internal_scene_conditions(scene_conditions):
    internal_scene_conditions = _zivid.SceneConditions()

    internal_scene_conditions.ambient_light = (
        _to_internal_scene_conditions_ambient_light(scene_conditions.ambient_light)
    )
    return internal_scene_conditions



"""Contains Application class."""

import _zivid
from zivid.camera import Camera


class Application:
    """Manager class for Zivid.

    When the first instance of this class is created it will initialize Zivid
    resources like camera management and GPU management.

    The resources will exist until release() is called, they will not be
    garbage collected even if the Application instance is. Subsequent instances
    of this class will refer to the already initialized resources.

    Calling release() on one instance of this class will invalidate all other
    instances of the class.

    This class can be used as a context manager to guarantee that resources are
    released deterministically. Note that this will also invalidate all other
    instances of this class.
    """

    def __init__(self):
        """Initialize application."""
        self.__impl = _zivid.Application()

    def __str__(self):
        return str(self.__impl)

    def create_file_camera(self, camera_file):
        """Create a virtual camera to simulate Zivid measurements by reading data from a file.

        An example file camera may be found among the Sample Data at zivid.com/downloads

        Args:
            camera_file: A pathlib.Path instance or a string specifying a Zivid File Camera (ZFC) file

        Returns:
            Zivid virtual Camera instance
        """
        return Camera(self.__impl.create_file_camera(str(camera_file)))

    def connect_camera(self, serial_number=None):
        """Connect to the next available Zivid camera.

        Args:
            serial_number: Optional serial number string for connecting to a specific camera

        Returns:
            Zivid Camera instance
        """
        if serial_number is not None:
            return Camera(self.__impl.connect_camera(serial_number))
        return Camera(self.__impl.connect_camera())

    def cameras(self):
        """Get a list of all cameras.

        Returns:
            A list of Camera including all physical cameras as well as virtual ones
                (e.g. cameras created by create_file_camera())
        """
        return [Camera(internal_camera) for internal_camera in self.__impl.cameras()]

    def release(self):
        """Release the underlying resources."""
        try:
            impl = self.__impl
        except AttributeError:
            pass
        else:
            impl.release()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.release()



from zivid.frame import Frame
from zivid.experimental.point_cloud_export.file_format import ZDF, PLY, XYZ, PCD

import _zivid


def export_frame(frame, file_format):
    """Save frame to a file.

    The file format is specified by the file_format argument. The file format can be ZDF, PLY, XYZ, or PCD.

    If the format is PCD, this function stores the ordered point cloud with a header that indicates an unordered point
    cloud. Since SDK 2.5, it is possible to export PCD with correct header by setting
    `Configuration/APIBreakingBugFixes/FileFormats/PCD/UseOrganizedFormat` in Config.yml file. See
    https://support.zivid.com/en/latest/reference-articles/point-cloud-structure-and-output-formats.html#organized-pcd-format.

    Args:
        frame: Frame to export.
        file_format: File format specification.

    Raises:
        TypeError: If frame is not a Frame.
        TypeError: If file_format is not a file format specification.
    """
    if not isinstance(frame, Frame):
        raise TypeError(
            "Unsupported type for argument frame. Got {}, expected {}".format(
                type(frame), Frame
            )
        )
    if not any(
        [
            isinstance(file_format, ZDF),
            isinstance(file_format, PLY),
            isinstance(file_format, XYZ),
            isinstance(file_format, PCD),
        ]
    ):
        raise TypeError(
            "Unsupported type for argument file_format. Got {}, expected {}".format(
                type(file_format),
                " or ".join([t.__name__ for t in [ZDF, PLY, XYZ, PCD]]),
            )
        )

    format_impl_attr = f"_{type(file_format).__name__}__impl"
    _zivid.point_cloud_export.export_frame(
        frame._Frame__impl,  # pylint: disable=protected-access
        getattr(file_format, format_impl_attr),
    )



"""This file imports all non protected classes, modules and packages from the current level."""

import zivid._version

__version__ = zivid._version.get_version(__name__)  # pylint: disable=protected-access

import zivid.firmware
import zivid.capture_assistant
import zivid.calibration
import zivid.presets
import zivid.projection

from zivid.application import Application
from zivid.camera import Camera
from zivid.camera_state import CameraState
from zivid.frame import Frame
from zivid.frame_2d import Frame2D
from zivid.frame_info import FrameInfo
from zivid.image import Image
from zivid.point_cloud import PointCloud
from zivid.sdk_version import SDKVersion
from zivid.settings import Settings
from zivid.settings2d import Settings2D
from zivid.camera_info import CameraInfo
from zivid.camera_intrinsics import CameraIntrinsics
from zivid.matrix4x4 import Matrix4x4
from zivid.network_configuration import NetworkConfiguration
from zivid.scene_conditions import SceneConditions



import pytest


def test_create_file_camera(application, file_camera_file):
    import zivid

    file_camera = application.create_file_camera(file_camera_file)
    assert file_camera
    assert isinstance(file_camera, zivid.camera.Camera)


@pytest.mark.physical_camera
def test_connect_camera(application):
    import zivid

    cam = application.connect_camera()
    assert cam
    assert isinstance(cam, zivid.camera.Camera)
    cam.release()


@pytest.mark.physical_camera
def test_connect_camera_serial_number(application):
    import zivid

    with application.connect_camera() as cam:
        serial_number = cam.info.serial_number

    with application.connect_camera(serial_number) as cam:
        assert cam
        assert isinstance(cam, zivid.camera.Camera)


def test_cameras_list_of_cameras(application):
    import zivid

    cameras = application.cameras()
    assert isinstance(cameras, list)
    for camera in cameras:
        assert isinstance(camera, zivid.Camera)


def test_cameras_one_camera(application, file_camera_file):
    orig_len = len(application.cameras())
    with application.create_file_camera(file_camera_file) as file_camera:
        assert file_camera
        cameras = application.cameras()
        assert len(cameras) == orig_len + 1
        assert file_camera in cameras


def test_to_string(application):
    string = str(application)
    assert string
    assert isinstance(string, str)



"""Module containing the Pose class.

This module should not be imported directly by end-user, but rather accessed through
the zivid.calibration module.
"""

import _zivid


class Pose:
    """Class representing a robot pose."""

    def __init__(self, transformation_matrix):
        """Construct a Pose object.

        Args:
            transformation_matrix:  A 4x4 array representing the pose
        """

        self.__impl = _zivid.calibration.Pose(transformation_matrix)

    def to_matrix(self):
        """Get the matrix representation of the pose.

        Returns:
            A 4x4 transformation matrix
        """
        return self.__impl.to_matrix()

    def __str__(self):
        return str(self.__impl)



"""Contains the Frame class."""

from pathlib import Path

import _zivid
from zivid.settings import _to_settings
from zivid.camera_info import _to_camera_info
from zivid.camera_state import _to_camera_state
from zivid.frame_info import _to_frame_info
from zivid.point_cloud import PointCloud
from zivid.frame_2d import Frame2D


class Frame:
    """A frame captured by a Zivid camera.

    Contains the point cloud (stored on compute device memory) as well as
    calibration data, settings and state used by the API at time of the frame
    capture. Use the point_cloud() method to access point cloud data.
    """

    def __init__(self, file_name):
        """Create a frame by loading data from a file.

        Args:
            file_name: A pathlib.Path instance or a string specifying a Zivid Data File (ZDF)

        Raises:
            TypeError: Unsupported type provided for file name
        """
        if isinstance(file_name, (str, Path)):
            self.__impl = _zivid.Frame(str(file_name))
        elif isinstance(file_name, _zivid.Frame):
            self.__impl = file_name
        else:
            raise TypeError(
                "Unsupported type for argument file_name. Got {}, expected {} or {}.".format(
                    type(file_name), str.__name__, Path.__name__
                )
            )

    def __str__(self):
        return str(self.__impl)

    def point_cloud(self):
        """Get the point cloud.

        See documentation/functions of zivid.PointCloud for instructions on how to
        retrieve point cloud data on various formats.

        Returns:
            A PointCloud instance
        """
        return PointCloud(self.__impl.point_cloud())

    def frame_2d(self):
        """Get 2D frame from 2D+3D frame.

        If the frame is the result of a 2D+3D capture, this method returns the 2D frame contained in the 2D+3D frame. In
        the case of a 3D-only capture, this method returns None.

        If the frame was captured by an SDK version prior to 2.14.0, then this method will return None.

        If the 2D frame is not yet available because the capture is still in-progress, then this method will block until
        acquisition of the entire 2D+3D capture is done. If you need to access the 2D frame before the 3D acquisition
        has finished, then it is required to do separate 2D and 3D captures.

        In a 2D+3D capture, the 2D color image and 3D point cloud may have different resolutions depending on the pixel
        sampling settings used. The 2D pixel sampling setting determines the resolution of the 2D color image whereas
        the 3D pixel sampling setting and the resampling setting determines the resolution of the 3D point cloud. The 2D
        color image returned in this 2D frame will always have the same resolution as the 2D color image that was
        captured. On the other hand, the point cloud colors will be sampled from the 2D color image to match the
        resolution of the 3D point cloud. The point cloud colors will always have a 1:1 correspondence with the 3D point
        cloud resolution. See `PointCloud` for more information.

        Returns:
            A Frame instance containing the 2D frame, or None if the frame was captured without 2D color or by an SDK
            version prior to 2.14.0.
        """
        return (
            Frame2D(self.__impl.frame_2d())
            if self.__impl.frame_2d() is not None
            else None
        )

    def save(self, file_path):
        """Save the frame to file. The file type is determined from the file extension.

        Args:
            file_path: A pathlib.Path instance or a string specifying destination

        """
        self.__impl.save(str(file_path))

    def load(self, file_path):
        """Load a frame from a Zivid data file.

        Args:
            file_path: A pathlib.Path instance or a string specifying a ZDF file to load
        """
        self.__impl.load(str(file_path))

    @property
    def settings(self):
        """Get the settings used to capture this frame.

        Returns:
            A Settings instance
        """
        return _to_settings(self.__impl.settings)

    @property
    def state(self):
        """Get the camera state data at the time of the frame capture.

        Returns:
            A CameraState instance
        """
        return _to_camera_state(self.__impl.state)

    @property
    def info(self):
        """Get information collected at the time of the frame capture.

        Returns:
            A FrameInfo instance
        """
        return _to_frame_info(self.__impl.info)

    @property
    def camera_info(self):
        """Get information about the camera used to capture the frame.

        Returns:
            A CameraInfo instance
        """
        return _to_camera_info(self.__impl.camera_info)

    def release(self):
        """Release the underlying resources."""
        try:
            impl = self.__impl
        except AttributeError:
            pass
        else:
            impl.release()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.release()

    def __del__(self):
        self.release()



import pytest


def test_is_up_to_date(shared_file_camera):
    from zivid.firmware import is_up_to_date

    assert is_up_to_date(shared_file_camera)


def test_update_without_callback(file_camera):
    import _zivid
    from zivid.firmware import update

    file_camera.disconnect()
    with pytest.raises(RuntimeError):
        update(file_camera)


def test_update_with_progress_callback(file_camera):
    import _zivid
    from zivid.firmware import update

    def callback(progress, description):
        print("{}% {}".format(progress, description), flush=True)

    file_camera.disconnect()

    with pytest.raises(RuntimeError):
        update(file_camera, progress_callback=callback)


def test_update_illegal_callback(file_camera):
    from zivid.firmware import update

    def illegal_callback():
        pass

    file_camera.disconnect()

    with pytest.raises(TypeError):
        update(file_camera, illegal_callback)



from datetime import datetime
import zivid


def test_time_stamp(frame_info):
    time_stamp = frame_info.time_stamp
    assert time_stamp
    assert isinstance(time_stamp, datetime)


def test_software_version(frame_info):
    software_version = frame_info.software_version
    assert software_version
    assert isinstance(software_version, zivid.frame_info.FrameInfo.SoftwareVersion)
    assert software_version.core
    assert isinstance(software_version.core, str)


def test_system_info(frame_info):
    system_info = frame_info.system_info
    assert isinstance(system_info, zivid.FrameInfo.SystemInfo)

    cpu = system_info.cpu
    assert isinstance(cpu, zivid.FrameInfo.SystemInfo.CPU)
    cpu_model = cpu.model
    assert cpu_model
    assert isinstance(cpu_model, str)

    compute_device = system_info.compute_device
    assert isinstance(compute_device, zivid.FrameInfo.SystemInfo.ComputeDevice)
    compute_model = compute_device.model
    assert compute_model
    assert isinstance(compute_model, str)
    compute_vendor = compute_device.vendor
    assert compute_vendor
    assert isinstance(compute_vendor, str)

    os_name = system_info.operating_system
    assert os_name
    assert isinstance(os_name, str)


def test_set_time_stamp(frame_info):
    assert isinstance(frame_info.time_stamp, datetime)
    assert isinstance(str(frame_info.time_stamp), str)
    new_time_stamp = datetime(1992, 2, 7)
    frame_info.time_stamp = new_time_stamp
    assert isinstance(str(frame_info.time_stamp), str)
    assert isinstance(frame_info.time_stamp, datetime)
    assert frame_info.time_stamp == new_time_stamp
